#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "chunks", "dtype", "fill_value"),
    [
        ((401, 100), (200, 10), "S10", "Hola!"),  # repeated string
        ((1020, 100), (200, 20), np.bool_, False),  # zeros
        ((1000, 99), (200, 20), np.int32, 1),  # ones
        ((799, 99), (20, 20), np.float64, np.nan),  # repeated float
    ],
)
def test_iterchunks_info(shape, chunks, dtype, fill_value):
    a = blosc2.full(shape, fill_value=fill_value, chunks=chunks, dtype=dtype)
    slice_ = (slice(0, chunks[0]), slice(0, chunks[1]))
    a[slice_] = 0  # introduce a zeroed chunk (another type of special value)

    for i, info in enumerate(a.iterchunks_info()):
        # print(info)
        assert info.nchunk == i
        if info.special == blosc2.SpecialValue.NOT_SPECIAL:
            assert info.cratio >= 10
        else:
            assert info.cratio >= 50



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
import os

import pytest

import blosc2


# This still needs to pass the '-s' flag to pytest to see the output but anyways
@pytest.fixture(scope="session", autouse=True)
def _setup_session():
    # This code will be executed before the test suite
    print()
    blosc2.print_versions()


@pytest.fixture(scope="session")
def c2sub_context():
    # You may use the URL and credentials for an already existing user
    # in a different Caterva2 subscriber.
    urlbase = os.environ.get("BLOSC_C2URLBASE", "https://demo.caterva2.net/")
    c2params = {"urlbase": urlbase, "username": None, "password": None}
    with blosc2.c2context(**c2params):
        yield c2params


# This is to avoid sporadic failures in the CI when reaching network,
# but this makes the tests to stuck in local.  Perhaps move this to
# every test module that needs it?
# def pytest_runtest_call(item):
#     try:
#         item.runtest()
#     except httpx.ConnectTimeout:
#         pytest.skip("Skipping test due to sporadic httpx.ConnectTimeout")
#     except httpx.ReadTimeout:
#         pytest.skip("Skipping test due to sporadic httpx.ReadTimeout")
#     except httpx.TimeoutException:
#         pytest.skip("Skipping test due to sporadic httpx.Timeout")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "new_shape", "chunks", "blocks", "fill_value"),
    [
        ((100, 1230), (200, 1230), (200, 100), (55, 3), b"0123"),
        ((23, 34), (23, 120), (20, 20), (10, 10), 1234),
        ((80, 51, 60), (80, 100, 100), (20, 10, 33), (6, 6, 26), 3.333),
    ],
)
def test_resize(shape, new_shape, chunks, blocks, fill_value):
    a = blosc2.full(shape, fill_value=fill_value, chunks=chunks, blocks=blocks)

    a.resize(new_shape)
    assert a.shape == new_shape

    slices = tuple(slice(s) for s in shape)
    for i in np.nditer(a[slices]):
        assert i == fill_value



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import os
from dataclasses import asdict, fields, replace

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("urlpath", "contiguous", "mode", "mmap_mode"),
    [
        (None, False, "r", None),
        (None, False, "w", None),
        (None, False, "a", None),
        (None, True, "r", None),
        (None, True, "w", None),
        (None, True, "a", None),
        ("b2frame", False, "r", None),
        ("b2frame", False, "w", None),
        ("b2frame", False, "a", None),
        ("b2frame", True, "r", None),
        ("b2frame", True, "w", None),
        ("b2frame", True, "a", None),
        ("b2frame", True, "r", "r"),
        ("b2frame", True, "w", "w+"),
        ("b2frame", True, "a", "w+"),  # r+ cannot be used here because the file does not exist
    ],
)
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks"),
    [
        (blosc2.CParams(codec=blosc2.Codec.LZ4, clevel=6, typesize=4), blosc2.DParams(), 0),
        ({"typesize": 4}, blosc2.DParams(nthreads=4), 1),
        (blosc2.CParams(splitmode=blosc2.SplitMode.ALWAYS_SPLIT, nthreads=5, typesize=4), {}, 5),
        ({"codec": blosc2.Codec.LZ4HC, "typesize": 4}, {}, 10),
    ],
)
def test_schunk_numpy(contiguous, urlpath, mode, mmap_mode, cparams, dparams, nchunks):
    storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath, mode=mode, mmap_mode=mmap_mode)
    blosc2.remove_urlpath(urlpath)

    chunk_len = 200 * 1000
    if mode != "r":
        schunk = blosc2.SChunk(chunksize=chunk_len * 4, storage=storage, cparams=cparams, dparams=dparams)

    else:
        with pytest.raises(
            ValueError, match="not specify a urlpath" if urlpath is None else "does not exist"
        ):
            blosc2.SChunk(chunksize=chunk_len * 4, storage=storage, cparams=cparams, dparams=dparams)

        # Create a schunk which we can read later
        storage2 = replace(
            storage,
            mode="w" if mmap_mode is None else None,
            mmap_mode="w+" if mmap_mode is not None else None,
        )
        schunk = blosc2.SChunk(chunksize=chunk_len * 4, storage=storage2, cparams=cparams, dparams=dparams)

    assert schunk.urlpath == urlpath
    assert schunk.contiguous == contiguous

    for i in range(nchunks):
        buffer = i * np.arange(chunk_len, dtype="int32")
        nchunks_ = schunk.append_data(buffer)
        assert nchunks_ == (i + 1)

    if mode == "r":
        if urlpath is not None:
            schunk = blosc2.SChunk(chunksize=chunk_len * 4, **asdict(storage))
        else:
            return
    assert schunk.nchunks == nchunks

    for i in range(nchunks):
        buffer = i * np.arange(chunk_len, dtype="int32")
        bytes_obj = buffer.tobytes()
        res = schunk.decompress_chunk(i)
        assert res == bytes_obj

        dest = np.empty(buffer.shape, buffer.dtype)
        schunk.decompress_chunk(i, dest)
        assert np.array_equal(buffer, dest)

        schunk.decompress_chunk(i, memoryview(dest))
        assert np.array_equal(buffer, dest)

        dest = bytearray(buffer)
        schunk.decompress_chunk(i, dest)
        assert dest == bytes_obj

    for i in range(nchunks):
        schunk.get_chunk(i)

    if nchunks >= 2:
        assert schunk.cratio > 1
        assert schunk.cratio == schunk.nbytes / schunk.cbytes
    assert schunk.nbytes >= nchunks * chunk_len * 4

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("mode_write", "mode_read", "mmap_mode_write", "mmap_mode_read"),
    [("w", "r", None, None), (None, None, "w+", "r")],
)
def test_schunk_ndarray(tmp_path, mode_write, mode_read, mmap_mode_write, mmap_mode_read):
    urlpath = tmp_path / "test.b2nd"

    data = np.arange(2 * 10, dtype="int32")
    blosc2.asarray(data, urlpath=urlpath, mode=mode_write, mmap_mode=mmap_mode_write)
    with pytest.raises(ValueError, match="Cannot open an NDArray as a SChunk"):
        blosc2.SChunk(mode=mode_read, mmap_mode=mmap_mode_read, urlpath=urlpath)


@pytest.mark.parametrize(
    ("urlpath", "contiguous", "mode", "mmap_mode"),
    [
        (None, False, "w", None),
        (None, True, "w", None),
        ("b2frame", False, "w", None),
        ("b2frame", True, "w", None),
        ("b2frame", True, None, "w+"),
    ],
)
@pytest.mark.parametrize(
    ("nbytes", "cparams", "dparams", "nchunks"),
    [
        (7, blosc2.CParams(codec=blosc2.Codec.LZ4, clevel=6, typesize=5), {}, 1),
        (641091, {"typesize": 3}, blosc2.DParams(nthreads=2), 1),
        (136, blosc2.CParams(typesize=1), blosc2.DParams(), 5),
        (1232, {"typesize": 8}, blosc2.dparams_dflts, 10),
    ],
)
def test_schunk(contiguous, urlpath, mode, mmap_mode, nbytes, cparams, dparams, nchunks):
    kwargs = {"contiguous": contiguous, "urlpath": urlpath, "cparams": cparams, "dparams": dparams}
    numpy_meta = {b"dtype": str(np.dtype(np.uint8))}
    test_meta = {b"lorem": 1234}
    meta = {"numpy": numpy_meta, "test": test_meta}
    blosc2.remove_urlpath(urlpath)

    schunk = blosc2.SChunk(chunksize=2 * nbytes, meta=meta, mode=mode, mmap_mode=mmap_mode, **kwargs)

    assert "numpy" in schunk.meta
    assert "error" not in schunk.meta
    assert schunk.meta["numpy"] == numpy_meta
    assert "test" in schunk.meta
    assert schunk.meta["test"] == test_meta
    test_meta = {b"lorem": 4231}
    schunk.meta["test"] = test_meta
    assert schunk.meta["test"] == test_meta

    for i in range(nchunks):
        bytes_obj = b"i " * nbytes
        nchunks_ = schunk.append_data(bytes_obj)
        assert nchunks_ == (i + 1)

    for i in range(nchunks):
        bytes_obj = b"i " * nbytes
        res = schunk.decompress_chunk(i)
        assert res == bytes_obj

        dest = bytearray(bytes_obj)
        schunk.decompress_chunk(i, dst=dest)
        assert dest == bytes_obj

    for i in range(nchunks):
        schunk.get_chunk(i)

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("urlpath", "contiguous", "mode", "mmap_mode"),
    [
        (None, False, "w", None),
        (None, True, "w", None),
        ("b2frame", False, "w", None),
        ("b2frame", True, "w", None),
        ("b2frame", True, None, "w+"),
    ],
)
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 4}, blosc2.DParams(), 1),
        ({"typesize": 4}, {"nthreads": 4}, 1),
        (blosc2.CParams(splitmode=blosc2.SplitMode.ALWAYS_SPLIT, nthreads=5, typesize=4), {}, 5),
        (blosc2.CParams(codec=blosc2.Codec.LZ4HC, typesize=4), blosc2.DParams(), 10),
    ],
)
@pytest.mark.parametrize("copy", [True, False])
def test_schunk_cframe(contiguous, urlpath, mode, mmap_mode, cparams, dparams, nchunks, copy):
    storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath, mode=mode, mmap_mode=mmap_mode)
    blosc2.remove_urlpath(urlpath)

    data = np.arange(200 * 1000 * nchunks, dtype="int32")
    schunk = blosc2.SChunk(
        chunksize=200 * 1000 * 4, data=data, **asdict(storage), cparams=cparams, dparams=dparams
    )

    cframe = schunk.to_cframe()
    schunk2 = blosc2.schunk_from_cframe(cframe, copy)
    cparams_dict = cparams if isinstance(cparams, dict) else asdict(cparams)
    if not os.getenv("BTUNE_TRADEOFF"):
        for key in cparams_dict:
            if key == "nthreads":
                continue
            if key == "blocksize" and cparams_dict[key] == 0:
                continue
            assert getattr(schunk2.cparams, key) == cparams_dict[key]

    data2 = np.empty(data.shape, dtype=data.dtype)
    schunk2.get_slice(out=data2)
    assert np.array_equal(data, data2)

    cframe = schunk.to_cframe()
    schunk3 = blosc2.schunk_from_cframe(cframe, copy)
    del schunk3
    # Check that we can still access the external cframe buffer
    _ = str(cframe)

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("cparams", "dparams", "new_cparams", "new_dparams"),
    [
        (
            blosc2.CParams(codec=blosc2.Codec.LZ4, clevel=6, typesize=4),
            {},
            blosc2.CParams(codec=blosc2.Codec.LZ4, clevel=6, typesize=4),
            blosc2.DParams(nthreads=4),
        ),
        (
            {"typesize": 4},
            blosc2.DParams(nthreads=4),
            blosc2.CParams(codec=blosc2.Codec.ZLIB, splitmode=blosc2.SplitMode.ALWAYS_SPLIT),
            blosc2.DParams(nthreads=1),
        ),
        (
            {"codec": blosc2.Codec.ZLIB, "splitmode": blosc2.SplitMode.ALWAYS_SPLIT},
            {},
            blosc2.CParams(
                splitmode=blosc2.SplitMode.ALWAYS_SPLIT,
                nthreads=5,
                typesize=4,
                filters=[blosc2.Filter.SHUFFLE, blosc2.Filter.TRUNC_PREC],
            ),
            blosc2.DParams(nthreads=16),
        ),
        (
            blosc2.CParams(codec=blosc2.Codec.LZ4HC, typesize=4),
            blosc2.DParams(),
            blosc2.CParams(filters=[blosc2.Filter.SHUFFLE, blosc2.Filter.TRUNC_PREC]),
            blosc2.DParams(nthreads=3),
        ),
    ],
)
def test_schunk_cdparams(cparams, dparams, new_cparams, new_dparams):
    kwargs = {"cparams": cparams, "dparams": dparams}

    chunk_len = 200 * 1000
    schunk = blosc2.SChunk(chunksize=chunk_len * 4, **kwargs)

    # Check cparams have been set correctly
    cparams_dict = cparams if isinstance(cparams, dict) else asdict(cparams)
    dparams_dict = dparams if isinstance(dparams, dict) else asdict(dparams)
    for key in cparams_dict:
        assert getattr(schunk.cparams, key) == cparams_dict[key]
    for key in dparams_dict:
        assert getattr(schunk.dparams, key) == dparams_dict[key]

    schunk.cparams = new_cparams
    schunk.dparams = new_dparams
    for field in fields(schunk.cparams):
        if field.name in ["filters", "filters_meta"]:
            assert getattr(schunk.cparams, field.name)[: len(getattr(new_cparams, field.name))] == getattr(
                new_cparams, field.name
            )
        else:
            assert getattr(schunk.cparams, field.name) == getattr(new_cparams, field.name)

    assert schunk.dparams.nthreads == new_dparams.nthreads



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "dtype", "urlpath"),
    [
        ((100, 123), "f4,f8", None),
        ((234, 125), "f4,(2,)f8", "test1.b2nd"),
        (80, [("f0", "<f4"), ("f1", "<f8")], "test2.b2nd"),
        ((40,), [("field 1", "<f4"), ("mamà", "<f8")], None),
        ((40,), [("field 1", "<f4"), ("mamà", "<f8")], "test3.b2nd"),
    ],
)
def test_scalar(shape, dtype, urlpath):
    blosc2.remove_urlpath(urlpath)

    a = blosc2.zeros(shape, dtype=dtype, urlpath=urlpath)
    b = np.zeros(shape=shape, dtype=dtype)
    assert np.array_equal(a[:], b)

    dtype = np.dtype(dtype)
    assert shape in (a.shape, a.shape[0])
    assert a.dtype == dtype
    assert a.schunk.typesize == dtype.itemsize
    assert a.shape == b.shape
    assert a.dtype == b.dtype

    if urlpath is not None:
        c = blosc2.open(urlpath)
        assert np.array_equal(c[:], b)
        assert c.shape == a.shape
        assert c.dtype == a.dtype
        assert c.schunk.typesize == dtype.itemsize
        assert c.shape == a.shape
        assert c.dtype == a.dtype

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize("contiguous", [True, False])
@pytest.mark.parametrize("urlpath", [None, "b2frame"])
@pytest.mark.parametrize(
    ("cparams", "dparams", "chunksize"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6}, {}, 40000),
        ({}, {"nthreads": 4}, 20000),
        ({"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 5}, {}, 20000),
        ({"codec": blosc2.Codec.LZ4HC, "typesize": 4}, {}, 40000),
    ],
)
def test_schunk_numpy(contiguous, urlpath, cparams, dparams, chunksize):
    kwargs = {"contiguous": contiguous, "urlpath": urlpath, "cparams": cparams, "dparams": dparams}
    blosc2.remove_urlpath(urlpath)
    num_elem = 20 * 1000
    nchunks = num_elem * 4 // chunksize + 1 if num_elem * 4 % chunksize != 0 else num_elem * 4 // chunksize
    data = np.arange(num_elem, dtype="int32")
    bytes_obj = data.tobytes()
    schunk = blosc2.SChunk(chunksize=chunksize, data=data, **kwargs)
    # Test properties
    assert len(schunk) == num_elem
    assert chunksize == schunk.chunksize
    assert chunksize / 4 == schunk.chunkshape
    assert cparams.get("blocksize", 0) == schunk.blocksize
    assert nchunks == schunk.nchunks
    assert num_elem * 4 == schunk.nbytes
    assert schunk.nbytes / schunk.cbytes == schunk.cratio
    assert schunk.typesize == 4

    for i in range(nchunks):
        start = i * chunksize
        np_start = start // 4
        if i == (nchunks - 1):
            end = len(bytes_obj)
        else:
            end = (i + 1) * chunksize
        np_end = end // 4
        res = schunk.decompress_chunk(i)
        assert res == bytes_obj[start:end]

        dest = np.empty(np_end - np_start, dtype=data.dtype)
        schunk.decompress_chunk(i, dest)
        assert np.array_equal(data[np_start:np_end], dest)

        schunk.decompress_chunk(i, memoryview(dest))
        assert np.array_equal(data[np_start:np_end], dest)

        dest = bytearray(data)
        schunk.decompress_chunk(i, dest[start:end])
        assert dest[start:end] == bytes_obj[start:end]

    for i in range(nchunks):
        schunk.get_chunk(i)

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize("contiguous", [True, False])
@pytest.mark.parametrize("urlpath", [None, "b2frame"])
@pytest.mark.parametrize(
    ("cparams", "dparams", "chunksize"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 1}, {}, 500),
        ({"typesize": 1}, {"nthreads": 4}, 500),
        ({"typesize": 1}, {}, 1000),
        ({"typesize": 1}, blosc2.dparams_dflts, 1000),
    ],
)
def test_schunk(contiguous, urlpath, cparams, dparams, chunksize):
    storage = {"contiguous": contiguous, "urlpath": urlpath}

    blosc2.remove_urlpath(urlpath)
    nrep = 1000
    nchunks = 5 * nrep // chunksize + 1 if nrep * 5 % chunksize != 0 else 5 * nrep // chunksize

    buffer = b"1234 " * nrep
    schunk = blosc2.SChunk(chunksize=chunksize, data=buffer, cparams=cparams, dparams=dparams, **storage)

    for i in range(nchunks):
        start = i * chunksize
        if i == (nchunks - 1):
            end = len(buffer)
        else:
            end = (i + 1) * chunksize
        bytes_obj = buffer[start:end]
        res = schunk.decompress_chunk(i)
        assert res == bytes_obj

        dest = bytearray(bytes_obj)
        schunk.decompress_chunk(i, dst=dest)
        assert dest == bytes_obj

    for i in range(nchunks):
        schunk.get_chunk(i)

    # Test properties
    assert chunksize == schunk.chunksize
    assert chunksize == schunk.chunkshape
    assert cparams.get("blocksize", 0) == schunk.blocksize
    assert nchunks == schunk.nchunks
    assert len(buffer) == schunk.nbytes
    assert schunk.nbytes / schunk.cbytes == schunk.cratio
    assert schunk.typesize == 1

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize("contiguous", [True, False])
@pytest.mark.parametrize("urlpath", [None, "b2frame"])
@pytest.mark.parametrize(
    ("cparams", "nitems"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 4}, 0),
        ({"typesize": 4}, 200 * 1000),
        ({"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 5, "typesize": 4}, 200 * 1000 * 2 + 17),
    ],
)
@pytest.mark.parametrize(
    ("special_value", "expected_value"),
    [
        (blosc2.SpecialValue.ZERO, 0),
        (blosc2.SpecialValue.NAN, np.nan),
        (blosc2.SpecialValue.UNINIT, 0),
        (blosc2.SpecialValue.VALUE, 34),
        (blosc2.SpecialValue.VALUE, np.pi),
        (blosc2.SpecialValue.VALUE, b"0123"),
        (blosc2.SpecialValue.VALUE, True),
    ],
)
def test_schunk_fill_special(contiguous, urlpath, cparams, nitems, special_value, expected_value):
    kwargs = {"contiguous": contiguous, "urlpath": urlpath, "cparams": cparams}
    blosc2.remove_urlpath(urlpath)

    chunk_len = 200 * 1000
    schunk = blosc2.SChunk(chunksize=chunk_len * 4, **kwargs)
    if special_value in [blosc2.SpecialValue.ZERO, blosc2.SpecialValue.NAN, blosc2.SpecialValue.UNINIT]:
        schunk.fill_special(nitems, special_value)
    else:
        schunk.fill_special(nitems, special_value, expected_value)
    assert len(schunk) == nitems

    if special_value != blosc2.SpecialValue.UNINIT:
        dtype = np.int32
        if isinstance(expected_value, float):
            dtype = np.float32
        elif isinstance(expected_value, bytes):
            dtype = np.dtype("|S" + str(len(expected_value)))
        array = np.full(nitems, expected_value, dtype=dtype)
        dest = np.empty(nitems, dtype=dtype)
        schunk.get_slice(out=dest)
        if dtype in [np.float32, np.float64]:
            np.testing.assert_allclose(dest, array)
        else:
            np.testing.assert_equal(dest, array)

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2

################# Automatic compute of optional chunks and blocks #################
# The exact outcome of these depends on many aspects, including CPUs cache sizes,
# so what is done here is mainly a shallow sanity check.  Enable the prints in
# case you want a detailed view of the guesses.
###################################################################################


@pytest.mark.parametrize("clevel", [0, 1, 5, 9])
@pytest.mark.parametrize("codec", [blosc2.Codec.BLOSCLZ, blosc2.Codec.ZSTD])
@pytest.mark.parametrize(
    "shape",
    [
        (1000, 1000),
        (10, 20, 30),
        (10, 30, 50, 10),
        (10, 10, 10, 10, 10),
    ],
)
@pytest.mark.parametrize("dtype", ["u1", "i4", "f8"])
def test_compute_chunks_blocks(clevel, codec, shape: tuple, dtype):
    cparams = blosc2.cparams_dflts.copy()
    cparams["clevel"] = clevel
    cparams["codec"] = codec
    cparams["typesize"] = np.dtype(dtype).itemsize
    if 0 in shape:
        # shapes with 0 should be reported as invalid
        with pytest.raises(ValueError):
            blosc2.compute_chunks_blocks(shape, **cparams)
        return
    else:
        chunks, blocks = blosc2.compute_chunks_blocks(shape, **cparams)
    # print(chunks, blocks)
    for dim, chunk, block in zip(shape, chunks, blocks, strict=True):
        assert dim >= chunk
        assert chunk >= block


@pytest.mark.parametrize(
    ("shape", "blocks"),
    [
        ((1000, 1000), (10, 10)),
        ((10, 10), (100, 100)),  # blocks can exceed shape if user wants to
        ((10, 20, 30), (1, 2, 3)),
        ((10, 20, 30), (1, 2, 31)),  # ditto, but even more weird blocks
        ((10, 30, 50, 10), (10, 30, 50, 10)),
        ((10, 10, 10, 10, 10), (10, 10, 10, 9, 10)),
        ((100, 10, 20, 100, 10), (10, 10, 10, 9, 10)),
        ((1000, 10, 20, 100, 10), (100, 10, 10, 90, 10)),
        ((1000, 10, 20, 100, 10), (100, 11, 10, 90, 10)),
    ],
)
def test_compute_chunks(shape: tuple, blocks: tuple):
    chunks, blocks = blosc2.compute_chunks_blocks(shape, blocks=blocks)
    # print(chunks, blocks)
    for dim, chunk, block in zip(shape, chunks, blocks, strict=True):
        assert dim >= chunk
        assert chunk >= block


# Invalid blocks
@pytest.mark.parametrize(
    ("shape", "blocks"),
    [
        ((1000, 1000), (0, 10)),  # zeros are not allowed
        ((10, 20, 30), (1, 2)),  # blocks need to have the same length as shape
    ],
)
def test_compute_chunks_except(shape: tuple, blocks: tuple):
    with pytest.raises(ValueError):
        blosc2.compute_chunks_blocks(shape, blocks=blocks)


@pytest.mark.parametrize(
    ("shape", "chunks"),
    [
        ((10, 10), (100, 100)),
        ((1000, 1000), (10, 10)),
        ((10, 20, 30), (1, 2, 3)),
        ((10, 30, 50, 10), (10, 30, 50, 10)),
        ((10, 10, 10, 10, 10), (10, 10, 10, 9, 10)),
        ((100, 10, 20, 100, 10), (10, 11, 10, 9, 10)),
        ((1000, 10, 20, 100, 10), (100, 11, 10, 90, 10)),
    ],
)
def test_compute_blocks(shape: tuple, chunks: tuple):
    chunks, blocks = blosc2.compute_chunks_blocks(shape, chunks=chunks)
    # print(chunks, blocks)
    for i in range(len(shape)):
        # assert shape[i] >= chunks[i]  # chunks can exceed shape if user wants to
        assert chunks[i] >= blocks[i]


@pytest.mark.parametrize(
    ("shape", "chunks"),
    [
        ((1000, 1000), (0, 10)),
        ((1000, 1000), (10,)),
    ],
)
def test_compute_blocks_except(shape: tuple, chunks: tuple):
    with pytest.raises(ValueError):
        blosc2.compute_chunks_blocks(shape, chunks=chunks)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Examples of using the jit decorator with expressions
# You can find benchmarks for this example in the bench/ndarray directory

import numpy as np

import blosc2


# Example 1: Basic usage of the jit decorator
@blosc2.jit
def expr_jit(a, b, c):
    # This function computes a boolean array where the condition is met
    return ((a**3 + np.sin(a * 2)) < c) & (b > 0)


# Create some sample data
a = blosc2.linspace(0, 1, 10 * 100, dtype="float32", shape=(10, 100))
b = blosc2.linspace(1, 2, 10 * 100, dtype="float32", shape=(10, 100))
c = blosc2.linspace(-10, 10, 10, dtype="float32", shape=(100,))

# Call the function with the jit decorator
result = expr_jit(a, b, c)
print(result[1, :10])

# Example 2: Using the jit decorator with an out parameter
out = blosc2.zeros((10, 100), dtype=np.bool_)


@blosc2.jit(out=out)
def expr_jit_out(a, b, c):
    # This function computes a boolean array and stores the result in the 'out' array
    return ((a**3 + np.sin(a * 2)) < c) & (b > 0)


# Call the function with the jit decorator and out parameter
result_out = expr_jit_out(a, b, c)
print(result_out[1, :10])
print(out[1, :10])  # The 'out' array should now contain the same result

# Example 3: Using the jit decorator with additional keyword arguments
cparams = blosc2.CParams(clevel=1, codec=blosc2.Codec.LZ4, filters=[blosc2.Filter.BITSHUFFLE])


@blosc2.jit(cparams=cparams)
def expr_jit_cparams(a, b, c):
    # This function computes a boolean array with custom compression parameters
    return ((a**3 + np.sin(a * 2)) < c) & (b > 0)


# Call the function with the jit decorator and custom parameters
result_cparams = expr_jit_cparams(a, b, c)
print(result_cparams[1, :10])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize("codec", blosc2.compressor_list())
def test_comp_info(codec):
    blosc2.clib_info(codec)
    blosc2.set_compressor(codec)
    assert codec.name.lower() == blosc2.get_compressor()

    arr = np.zeros(1_000_000, dtype="V8")
    src = blosc2.compress2(arr)
    nbytes, cbytes, blocksize = blosc2.get_cbuffer_sizes(src)
    assert nbytes == arr.size * arr.dtype.itemsize
    assert cbytes == blosc2.MAX_OVERHEAD
    # When raising the next limit when this would fail in the future, one should raise the SIZE too
    assert blocksize <= 2**23
    blosc2.print_versions()



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import io
import pprint
from textwrap import TextWrapper


def info_text_report_(items: list) -> str:
    with io.StringIO() as buf:
        print(items, file=buf)
        return buf.getvalue()


def info_text_report(items: list) -> str:
    keys = [k for k, v in items]
    max_key_len = max(len(k) for k in keys)
    report = ""
    for k, v in items:
        if isinstance(v, dict):
            # rich way, this is disabled because it doesn't work well in the notebooks
            # with io.StringIO() as buf:
            #     v_sorted = {k: val for k, val in sorted(v.items())}
            #     rich.print(v_sorted, file=buf)
            #     str_v = buf.getvalue()[:-1]  # remove the trailing \n
            # text = k.ljust(max_key_len) + " : " + str_v
            # pprint way
            text = k.ljust(max_key_len) + " : " + pprint.pformat(v)
        else:
            wrapper = TextWrapper(
                width=96,
                initial_indent=k.ljust(max_key_len) + " : ",
                subsequent_indent=" " * max_key_len + " : ",
            )
            text = wrapper.fill(str(v))
        report += text + "\n"
    return report


def info_html_report(items: list) -> str:
    report = '<table class="NDArray-info">'
    report += "<tbody>"
    for k, v in items:
        report += f'<tr><th style="text-align: left">{k}</th><td style="text-align: left">{v}</td></tr>'
    report += "</tbody>"
    report += "</table>"
    return report


class InfoReporter:
    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        items = self.obj.info_items
        return info_text_report(items)

    def _repr_html_(self):
        items = self.obj.info_items
        return info_html_report(items)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


"""
Small benchmark that exercises packaging of arrays larger than 2 GB.
"""

import time

import numpy as np

import blosc2

NREP = 1
N = int(4e8 - 2**27)  # larger than 2 GB
Nexp = np.log10(N)

print(f"Creating NumPy array with {float(N):.3g} int64 elements...")
in_ = np.arange(N, dtype=np.int64)

if __name__ == "__main__":
    cparams = {
        "codec": blosc2.Codec.BLOSCLZ,
        "clevel": 9,
        # "filters": [blosc2.Filter.NOFILTER] * 4 + [blosc2.Filter.SHUFFLE, blosc2.Filter.BYTEDELTA],
        # "filters_meta": [0] * 6,
        # "splitmode": blosc2.SplitMode.NEVER_SPLIT,
    }
    print(f"Storing with {cparams=}")

    c = None
    ctic = time.time()
    for _i in range(NREP):
        c = blosc2.pack_tensor(in_, cparams=cparams)
    ctoc = time.time()
    tc = (ctoc - ctic) / NREP
    print(
        f"  Time for pack_tensor:   {tc:.3f} ({(N * 8 / tc) / 2**30:.2f} GB/s)) ",
    )
    print(f"\tcr: {in_.size * in_.dtype.itemsize * 1.0 / len(c):5.1f}x")

    out = None
    dtic = time.time()
    for _i in range(NREP):
        out = blosc2.unpack_tensor(c)
    dtoc = time.time()

    td = (dtoc - dtic) / NREP
    print(
        f"  Time for unpack_tensor:   {td:.3f} s ({(N * 8 / td) / 2**30:.2f} GB/s)) ",
    )
    assert np.array_equal(in_, out)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
import pathlib

import numexpr as ne
import numpy as np
import pytest

import blosc2

pytestmark = pytest.mark.network

NITEMS_SMALL = 1_000
ROOT = "b2tests"
DIR = "expr/"


def get_arrays(shape, chunks_blocks):
    dtype = np.float64
    nelems = np.prod(shape)
    na1 = np.linspace(0, 10, nelems, dtype=dtype).reshape(shape)
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a1-{shape}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    a1 = blosc2.C2Array(path)
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a2-{shape}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    a2 = blosc2.C2Array(path)
    # Let other operands be local, on-disk NDArray copies
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a3-{shape}d.b2nd"
    a3 = blosc2.asarray(a2, urlpath=urlpath, mode="w")
    urlpath = f"ds-0-10-linspace-{dtype.__name__}-{chunks_blocks}-a4-{shape}d.b2nd"
    a4 = a3.copy(urlpath=urlpath, mode="w")
    assert isinstance(a1, blosc2.C2Array)
    assert isinstance(a2, blosc2.C2Array)
    assert isinstance(a3, blosc2.NDArray)
    assert isinstance(a4, blosc2.NDArray)
    return a1, a2, a3, a4, na1, np.copy(na1), np.copy(na1), np.copy(na1)


@pytest.mark.parametrize(
    "chunks_blocks",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_simple(chunks_blocks, c2sub_context):
    shape = (60, 60)
    a1, a2, a3, a4, na1, na2, na3, na4 = get_arrays(shape, chunks_blocks)

    # Slice
    sl = slice(10)
    expr = a1 + a3
    nres = ne.evaluate("na1 + na3")
    res = expr.compute(item=sl)
    np.testing.assert_allclose(res[:], nres[sl])

    # All
    res = expr.compute()
    np.testing.assert_allclose(res[:], nres)


def test_simple_getitem(c2sub_context):
    shape = (NITEMS_SMALL,)
    chunks_blocks = "default"
    a1, a2, a3, a4, na1, na2, na3, na4 = get_arrays(shape, chunks_blocks)
    expr = a1 + a2 - a3 * a4
    nres = ne.evaluate("na1 + na2 - na3 * na4")

    # slice
    sl = slice(10)
    res = expr[sl]
    np.testing.assert_allclose(res, nres[sl])
    # all
    res = expr[:]
    np.testing.assert_allclose(res, nres)


# Add more test functions to test different aspects of the code
@pytest.mark.parametrize(
    "chunks_blocks",
    [
        (True, False),
        (False, False),
    ],
)
def test_ixxx(chunks_blocks, c2sub_context):
    shape = (60, 60)
    a1, a2, a3, a4, na1, na2, na3, na4 = get_arrays(shape, chunks_blocks)
    expr = a1**3 + a2**2 + a3**3 - a4 + 3
    expr += 5  # __iadd__
    expr /= 7  # __itruediv__
    expr **= 2.3  # __ipow__
    res = expr.compute()
    nres = ne.evaluate("(((na1 ** 3 + na2 ** 2 + na3 ** 3 - na4 + 3) + 5) / 7) ** 2.3")
    np.testing.assert_allclose(res[:], nres)


def test_complex(c2sub_context):
    shape = (NITEMS_SMALL,)
    chunks_blocks = "default"
    a1, a2, a3, a4, na1, na2, na3, na4 = get_arrays(shape, chunks_blocks)
    expr = blosc2.tan(a1) * blosc2.sin(a2) + (blosc2.sqrt(a4) * 2)
    expr += 2
    nres = ne.evaluate("tan(na1) * sin(na2) + (sqrt(na4) * 2) + 2")
    # eval
    res = expr.compute()
    np.testing.assert_allclose(res[:], nres)
    # __getitem__
    res = expr[:]
    np.testing.assert_allclose(res, nres)
    # slice
    sl = slice(10)
    res = expr[sl]
    np.testing.assert_allclose(res, nres[sl])


# Test expr with remote & local operands
@pytest.mark.parametrize(
    "chunks_blocks",
    [
        pytest.param((True, True), marks=pytest.mark.heavy),
        pytest.param((True, False), marks=pytest.mark.heavy),
        pytest.param((False, True), marks=pytest.mark.heavy),
        (False, False),
    ],
)
def test_mix_operands(chunks_blocks, c2sub_context):
    shape = (60, 60)
    a1, a2, a3, a4, na1, na2, na3, na4 = get_arrays(shape, chunks_blocks)
    b1 = blosc2.asarray(na1, chunks=a1.chunks, blocks=a1.blocks)
    b3 = blosc2.asarray(na3, chunks=a3.chunks, blocks=a3.blocks)

    expr = a1 + b1
    nres = ne.evaluate("na1 + na1")
    np.testing.assert_allclose(expr[:], nres)
    np.testing.assert_allclose(expr.compute()[:], nres)

    expr = a1 + b3
    nres = ne.evaluate("na1 + na3")
    np.testing.assert_allclose(expr[:], nres)
    np.testing.assert_allclose(expr.compute()[:], nres)

    expr = a1 + b1 + a2 + b3
    nres = ne.evaluate("na1 + na1 + na2 + na3")
    np.testing.assert_allclose(expr[:], nres)
    np.testing.assert_allclose(expr.compute()[:], nres)

    expr = a1 + a2 + b1 + b3
    nres = ne.evaluate("na1 + na2 + na1 + na3")
    np.testing.assert_allclose(expr[:], nres)
    np.testing.assert_allclose(expr.compute()[:], nres)

    # TODO: fix this
    # expr = a1 + na1 * b3
    # print(type(expr))
    # print("expression: ", expr.expression)
    # nres = ne.evaluate("na1 + na1 * na3")
    # np.testing.assert_allclose(expr[:], nres)
    # np.testing.assert_allclose(expr.compute()[:], nres)


# Tests related with save method
def test_save(c2sub_context):
    shape = (60, 60)
    tol = 1e-17
    a1, a2, a3, a4, na1, na2, na3, na4 = get_arrays(shape, (False, True))

    expr = a1 * a2 + a3 - a4 * 3
    nres = ne.evaluate("na1 * na2 + na3 - na4 * 3")

    res = expr.compute()
    assert res.dtype == np.float64
    np.testing.assert_allclose(res[:], nres, rtol=tol, atol=tol)

    urlpath = "expr.b2nd"
    expr.save(urlpath=urlpath, mode="w")
    ops = [a1, a2, a3, a4]
    for op in ops:
        del op
    del expr
    expr = blosc2.open(urlpath)
    res = expr.compute()
    assert res.dtype == np.float64
    np.testing.assert_allclose(res[:], nres, rtol=tol, atol=tol)
    # Test getitem
    np.testing.assert_allclose(expr[:], nres, rtol=tol, atol=tol)

    blosc2.remove_urlpath(urlpath)


@pytest.fixture(
    params=[
        ((2, 5), (5,)),
        pytest.param(((2, 1), (5,)), marks=pytest.mark.heavy),
        pytest.param(((2, 5, 3), (5, 1)), marks=pytest.mark.heavy),
        ((2, 1, 3), (5, 3)),
        pytest.param(((2, 5, 3, 2), (5, 3, 1)), marks=pytest.mark.heavy),
        ((2, 5, 3, 2), (5, 1, 2)),
        pytest.param(((2, 5, 3, 2, 2), (5, 3, 2, 2)), marks=pytest.mark.heavy),
    ]
)
def broadcast_shape(request):
    return request.param


@pytest.fixture
def broadcast_fixture(broadcast_shape, c2sub_context):
    shape1, shape2 = broadcast_shape
    dtype = np.float64
    na1 = np.linspace(0, 1, np.prod(shape1), dtype=dtype).reshape(shape1)
    na2 = np.linspace(1, 2, np.prod(shape2), dtype=dtype).reshape(shape2)
    urlpath = f"ds-0-1-linspace-{dtype.__name__}-b1-{shape1}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    b1 = blosc2.C2Array(path)
    urlpath = f"ds-1-2-linspace-{dtype.__name__}-b2-{shape2}d.b2nd"
    path = pathlib.Path(f"{ROOT}/{DIR + urlpath}").as_posix()
    b2 = blosc2.C2Array(path)

    return b1, b2, na1, na2


def test_broadcasting(broadcast_fixture):
    a1, a2, na1, na2 = broadcast_fixture
    expr1 = a1 + a2
    assert expr1.shape == np.broadcast_shapes(a1.shape, a2.shape)
    expr2 = a1 * a2 + 1
    assert expr2.shape == np.broadcast_shapes(a1.shape, a2.shape)
    expr = expr1 - expr2
    assert expr.shape == np.broadcast_shapes(a1.shape, a2.shape)
    nres = ne.evaluate("na1 + na2 - (na1 * na2 + 1)")
    res = expr.compute()
    np.testing.assert_allclose(res[:], nres)
    res = expr[:]
    np.testing.assert_allclose(res, nres)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#
# This example can only be run if blosc2-btune is installed. You can
# get it from https://pypi.org/project/blosc2-btune/
# For more info on this tuner plugin see
# https://github.com/Blosc/blosc2_btune/blob/main/README.md
#######################################################################

import blosc2_btune
import numpy as np

import blosc2

nchunks = 10
# Set the compression and decompression parameters, use BTUNE tuner
cparams = blosc2.CParams(codec=blosc2.Codec.LZ4HC, typesize=4, tuner=blosc2.Tuner.BTUNE)
dparams = blosc2.DParams()
contiguous = True
urlpath = "filename"

storage = blosc2.Storage(contiguous=contiguous, urlpath=urlpath, mode="a")
blosc2.remove_urlpath(urlpath)

# Set the Btune configuration to use
btune_conf = {"tradeoff": 0.3, "perf_mode": blosc2_btune.PerformanceMode.DECOMP}
blosc2_btune.set_params_defaults(**btune_conf)

# Create the SChunk
data = np.arange(200 * 1000 * nchunks)
schunk = blosc2.SChunk(
    chunksize=200 * 1000 * 4, data=data, cparams=cparams, dparams=dparams, storage=storage
)

# Check data can be retrieved correctly
data2 = np.empty(data.shape, dtype=data.dtype)
schunk.get_slice(out=data2)
assert np.array_equal(data, data2)

blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# This shows how to evaluate expressions with NDArray instances as operands.

import numpy as np

import blosc2

shape = (10, 1, 2)

# Create a NDArray from a NumPy array
npa = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
npb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
npc = npa**2 + npb**2 + 2 * npa * npb + 1

a = blosc2.asarray(npa, urlpath="a.b2nd", mode="w")
b = blosc2.asarray(npb, urlpath="b.b2nd", mode="w")

# Get a LazyExpr instance
c = a**2 + b**2 + 2 * a * b + 1
c.save(urlpath="c.b2nd")
c = blosc2.open("c.b2nd")
# Evaluate: output is a NDArray
d = blosc2.lazyexpr("a + c.sum() + a.std()", operands={"a": a, "c": c})
d.save(urlpath="lazy-d.b2nd")

# Load the expression from disk
d = blosc2.open("lazy-d.b2nd")
print(f"Expression: {d}")
assert isinstance(d, blosc2.LazyExpr)
e = d.compute()
assert isinstance(e, blosc2.NDArray)
sum = e[()]
print("Reduction with Blosc2:\n", sum[1])
npsum = npa + np.sum(npc) + np.std(npa)
print("Reduction with NumPy:\n", npsum[1])
assert np.allclose(sum, npsum)

print("NDArray expression evaluated correctly in-memory!")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


# A simple example using the pack and unpack functions
import numpy as np

import blosc2

a = np.array(["å", "ç", "ø"])
parray = blosc2.pack(a, 9)
a2 = blosc2.unpack(parray)
assert np.all(a == a2)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


# A simple example using the save_tensor and load_tensor functions

import numpy as np

import blosc2

a = np.arange(1_000_000)

file_size = blosc2.save_tensor(a, "save_tensor.bl2", mode="w")
print("Length of saved tensor in file (bytes):", file_size)

a2 = blosc2.load_tensor("save_tensor.bl2")
assert np.all(a == a2)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# This shows how to evaluate expressions with NDArray instances as operands.

import numpy as np

import blosc2

shape = (50, 50)

# Create a NDArray from a NumPy array
npa = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
npb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
npc = npa**2 + npb**2 + 2 * npa * npb + 1

a = blosc2.asarray(npa)
b = blosc2.asarray(npb)

# Get a LazyExpr instance
c = a**2 + b**2 + 2 * a * b + 1
# Evaluate: output is a NDArray
d = c.compute()
# Check
assert isinstance(d, blosc2.NDArray)
assert np.allclose(d[:], npc)

# Evaluate the whole slice: output is a NumPy array
npd = c[:]
# Check
assert isinstance(npd, np.ndarray)
assert np.allclose(npd, npc)

# Evaluate a partial slice: output is a NumPy array
npd = c[1:10]
# Check
assert isinstance(npd, np.ndarray)
assert np.allclose(npd, npc[1:10])

print("NDArray expression evaluated correctly in-memory!")

# Now, evaluate the expression from operands in disk
# TODO: when doing a copy, mode should be 'w' by default?
da = a.copy(urlpath="a.b2nd", mode="w")
db = b.copy(urlpath="b.b2nd", mode="w")

# Get a LazyExpr instance
(da**2 + db**2 + 2 * da * db + 1).save(urlpath="c.b2nd")
dc = blosc2.open("c.b2nd")

# Evaluate: output is a NDArray
dc2 = dc.compute()
# Check
assert isinstance(dc2, blosc2.NDArray)
assert np.allclose(dc2[:], npc)
print("NDArray expression evaluated correctly on-disk!")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# This shows how to evaluate expressions in combination with the where() functionality.

import numpy as np

import blosc2

shape = (50, 50)
chunks = (10, 10)
blocks = (5, 5)

# Create a structured NumPy array
npa = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
npb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
npc = npa**2 + npb**2 > 2 * npa * npb + 1
nps = np.empty(shape, dtype=[("a", npa.dtype), ("b", npb.dtype)])
nps["a"] = npa
nps["b"] = npb

s = blosc2.asarray(nps, chunks=chunks, blocks=blocks)
a = blosc2.NDField(s, "a")
b = blosc2.NDField(s, "b")

# Get a LazyExpr instance
c = a**2 + b**2 > 2 * a * b + 1


# Simple where() method
d = c.where(0, 1)
# print(d[:])
np.testing.assert_allclose(d[:], np.where(npc, 0, 1))

d = blosc2.where(c, 0, 1)
# print(d[:])
np.testing.assert_allclose(d[:], np.where(npc, 0, 1))

d = blosc2.lazyexpr(c, where=(0, 1))
# print(d[:])
np.testing.assert_allclose(d[:], np.where(npc, 0, 1))


# Not sure if a decorator like this is a good idea, but it works
@blosc2.lazywhere(0, 1)
def myexpr(a, b):
    return a**2 + b**2 > 2 * a * b + 1


d = myexpr(a, b)
# print(d[:])
np.testing.assert_allclose(d[:], np.where(npc, 0, 1))

# where accepts only a single `x` parameter (not directly supported by NumPy)
d = c.where(s)
npd = d[:]
# print(npd)
np.testing.assert_allclose(npd["a"], nps[npc]["a"])
np.testing.assert_allclose(npd["b"], nps[npc]["b"])


# Decorator version
@blosc2.lazywhere(s)
def myexpr2(a, b):
    return a**2 + b**2 > 2 * a * b + 1


d = myexpr2(a, b)
npd = d[:]
# print(npd)
np.testing.assert_allclose(npd["a"], nps[npc]["a"])
np.testing.assert_allclose(npd["b"], nps[npc]["b"])


# # TODO: Test with no parameters
# d = c.where()
# print(d[:])
# np.testing.assert_allclose(d[:], npc.nonzero())

# NDArray.__getitem__ with LazyExpr (converted into c.where(s) behind the scenes)
d = s[a**2 + b**2 > 2 * a * b + 1]
npd = d[:]
# print(npd)
np.testing.assert_allclose(npd["a"], nps[npc]["a"])
np.testing.assert_allclose(npd["b"], nps[npc]["b"])

# NDArray.__getitem__ with a string expression
d = s["a**2 + b**2 > 2 * a * b + 1"]
npd = d[:]
print(npd)
np.testing.assert_allclose(npd["a"], nps[npc]["a"])
np.testing.assert_allclose(npd["b"], nps[npc]["b"])

# Combined with reductions
d = blosc2.where(c, 0, 1).sum(axis=1)
print(d[...])
np.testing.assert_allclose(d[...], np.where(npc, 0, 1).sum(axis=1))

print("blosc2.where is working correctly!")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import numpy as np
import pytest

import blosc2

random = np.random.default_rng()


@pytest.mark.parametrize("gil", [True, False])
@pytest.mark.parametrize(
    ("obj", "cparams", "dparams"),
    [
        (random.integers(0, 10, 10), {"cparams": blosc2.CParams(codec=blosc2.Codec.LZ4, clevel=6)}, {}),
        (
            np.arange(10, dtype="float32"),
            # Select an absolute precision of 10 bits in mantissa
            {
                "cparams": blosc2.CParams(
                    filters=[blosc2.Filter.TRUNC_PREC, blosc2.Filter.BITSHUFFLE],
                    filters_meta=[10, 0],
                    typesize=4,
                )
            },
            {"dparams": blosc2.DParams(nthreads=4)},
        ),
        (
            np.arange(10, dtype="float32"),
            # Do a reduction of precision of 10 bits in mantissa
            {
                "cparams": {
                    "filters": [blosc2.Filter.TRUNC_PREC, blosc2.Filter.BITSHUFFLE],
                    "filters_meta": [-10, 0],
                    "typesize": 4,
                },
            },
            {"nthreads": 4},
        ),
        (
            random.integers(0, 1000, 1000, endpoint=True),
            {"cparams": blosc2.CParams(splitmode=blosc2.SplitMode.ALWAYS_SPLIT, nthreads=5, typesize=4)},
            {"dparams": blosc2.DParams()},
        ),
        (
            np.arange(45, dtype=np.float64),
            {"cparams": blosc2.CParams(codec=blosc2.Codec.LZ4HC, typesize=4)},
            {},
        ),
        (np.arange(50, dtype=np.int64), {"typesize": 4}, {"dparams": blosc2.dparams_dflts}),
    ],
)
def test_compress2_numpy(obj, cparams, dparams, gil):
    blosc2.set_releasegil(gil)
    bytes_obj = obj.tobytes()
    c = blosc2.compress2(obj, **cparams)

    dest = bytearray(obj)
    blosc2.decompress2(c, dst=dest, **dparams)
    assert dest == bytes_obj

    dest2 = np.empty(obj.shape, obj.dtype)
    blosc2.decompress2(c, dst=dest2, **dparams)
    assert np.array_equal(dest2, obj)

    dest3 = blosc2.decompress2(c, **dparams)
    assert dest3 == bytes_obj

    dest4 = np.empty(obj.shape, obj.dtype)
    blosc2.decompress2(c, dst=memoryview(dest4), **dparams)
    assert np.array_equal(dest4, obj)


@pytest.mark.parametrize("gil", [True, False])
@pytest.mark.parametrize(
    ("obj", "cparams", "dparams"),
    [
        (
            random.integers(0, 10, 10, dtype=np.int64),
            {"codec": blosc2.Codec.LZ4, "clevel": 6, "filters_meta": [-50]},
            {},
        ),
        (
            np.arange(10, dtype="int32"),
            {"filters_meta": [-20]},
            {"nthreads": 4},
        ),
        (np.arange(45, dtype=np.int16), {"codec": blosc2.Codec.LZ4HC, "filters_meta": [-10]}, {}),
        (np.arange(50, dtype=np.int8), {"filters_meta": [-5]}, blosc2.dparams_dflts),
    ],
)
def test_compress2_int_trunc(obj, cparams, dparams, gil):
    blosc2.set_releasegil(gil)
    cparams["filters"] = [blosc2.Filter.INT_TRUNC]
    cparams["typesize"] = obj.dtype.itemsize
    c = blosc2.compress2(obj, **cparams)

    dest = np.empty(obj.shape, obj.dtype)
    blosc2.decompress2(c, dst=dest, **dparams)

    for i in range(obj.shape[0]):
        assert (obj[i] - dest[i]) <= (2 ** ((-1) * cparams["filters_meta"][0]))


@pytest.mark.parametrize("gil", [True, False])
@pytest.mark.parametrize(
    ("nbytes", "cparams", "dparams"),
    [
        (7, {"codec": blosc2.Codec.LZ4, "clevel": 6, "typesize": 1}, {}),
        (641091, {"typesize": 1}, {"nthreads": 4}),
        (136, {"typesize": 1}, {}),
        (1231, {"typesize": 4}, blosc2.dparams_dflts),
    ],
)
def test_compress2(nbytes, cparams, dparams, gil):
    blosc2.set_releasegil(gil)
    bytes_obj = b" " * nbytes
    c = blosc2.compress2(bytes_obj, **cparams)

    dest = bytearray(bytes_obj)
    blosc2.decompress2(c, dst=dest, **dparams)
    assert dest == bytes_obj

    dest2 = blosc2.decompress2(c, **dparams)
    assert dest2 == bytes_obj

    dest3 = bytearray(bytes_obj)
    blosc2.decompress2(np.array([c]), dst=dest3, **dparams)
    assert dest3 == bytes_obj


@pytest.mark.parametrize("gil", [True, False])
@pytest.mark.parametrize(
    ("object", "cparams", "dparams"),
    [(np.arange(0), {"codec": blosc2.Codec.LZ4, "clevel": 6}, {}), (b"", {}, {"nthreads": 3})],
)
def test_raise_error(object, cparams, dparams, gil):
    blosc2.set_releasegil(gil)
    c = blosc2.compress2(object, **cparams, **dparams)

    dest = bytearray(object)
    with pytest.raises(ValueError):
        blosc2.decompress2(c, dst=dest)

    dest3 = blosc2.decompress2(c)
    if isinstance(object, bytes):
        assert dest3 == object
    else:
        assert dest3 == object.tobytes()

    dest5 = bytearray(object)
    with pytest.raises(ValueError):
        blosc2.decompress2(np.array([c]), dst=dest5)

    with pytest.raises(ValueError):
        blosc2.decompress2(b"")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Fill an SChunk with a filler decorator

import numpy as np

import blosc2

nchunks = 3
chunk_len = 200 * 1000
schunk_dtype = np.dtype(np.float64)

# Set the compression parameters. We need nthreads=1 for this example.
cparams = blosc2.CParams(typesize=schunk_dtype.itemsize, nthreads=1)

# Create empty SChunk
schunk = blosc2.SChunk(chunksize=chunk_len * schunk_dtype.itemsize, cparams=cparams)

# Create operands (can be a SChunk, numpy.ndarray or Python scalar)
op_dtype = np.dtype(np.int32)
data = np.full(chunk_len * nchunks, 1234, dtype=op_dtype)
schunk_op = blosc2.SChunk(chunksize=chunk_len * op_dtype.itemsize, data=data)
op2_dtype = np.dtype(np.float32)
nparray_op = np.arange(0, chunk_len * nchunks, dtype=op2_dtype)
py_scalar = np.e


# Set filler with decorator
@schunk.filler(((schunk_op, op_dtype), (nparray_op, op2_dtype), (py_scalar, np.float32)), schunk_dtype)
def filler(inputs_tuple, output, offset):
    output[:] = inputs_tuple[0] - inputs_tuple[1] * inputs_tuple[2]


# Check that SChunk has been filled correctly
out = np.empty(chunk_len * nchunks, dtype=schunk_dtype)
schunk.get_slice(0, chunk_len * nchunks, out=out)

res = np.empty(data.shape, dtype=schunk_dtype)
filler((data, nparray_op, py_scalar), res, None)
np.testing.assert_allclose(out, res)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Create an empty array with different compression parameters and set some values on it

import blosc2

cparams = blosc2.CParams(
    codec=blosc2.Codec.LZ4,
    clevel=5,
    nthreads=4,
    filters=[blosc2.Filter.DELTA, blosc2.Filter.TRUNC_PREC, blosc2.Filter.BITSHUFFLE],
    filters_meta=[0, 3, 0],  # keep just 3 bits in mantissa
)
a = blosc2.empty(shape=(40, 401), blocks=(6, 26), dtype="f8", cparams=cparams)

a[...] = 222
print(a.info)

print(a[:, 0])  # note the truncation filter at work



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Resizing an array is simple (and efficient too)

import blosc2

a = blosc2.full((4, 4), fill_value=9)
a.resize((5, 7))
a[3:5, 2:7] = 8
print(a[:])



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


"""
Packaging tensors (PyTorch, TensorFlow) larger than 2 GB.
"""

import io
import sys
import time

import numpy as np
import tensorflow as tf
import torch

import blosc2

NREP = 1
# N = int(5e8 + 2**27)  # larger than 2 GB
# Using tensors > 2 GB makes tensorflow serialization to raise this error:
# [libprotobuf FATAL google/protobuf/io/coded_stream.cc:831] CHECK failed: overrun <= kSlopBytes:
N = int(1e8)

store = True
if len(sys.argv) > 1:
    store = True

# blosc2.set_nthreads(8)

print(f"Creating NumPy array with {float(N):.3g} float32 elements...")
# in_ = np.arange(N, dtype=np.float32)
in_ = np.linspace(0, 1, N, dtype=np.float32)

if store:
    tt = tf.constant(in_)
    th = torch.from_numpy(in_)

    # Standard TensorFlow serialization
    c = None
    ctic = time.time()
    for _i in range(NREP):
        c = tf.io.serialize_tensor(tt).numpy()
    ctoc = time.time()
    tc = (ctoc - ctic) / NREP
    print(
        f"  Time for tensorflow (tf.io.serialize):\t{tc:.3f} s ({(N * 8 / tc) / 2**30:.2f} GB/s)) ",
        end="",
    )
    print(f"\tcr: {in_.size * in_.dtype.itemsize * 1.0 / len(c):5.1f}x")

    with open("serialize_tensorflow.bin", "wb") as f:
        f.write(c)

    # Standard PyTorch serialization
    c = None
    buff = io.BytesIO()
    ctic = time.time()
    for _i in range(NREP):
        torch.save(th, buff)
    ctoc = time.time()
    tc = (ctoc - ctic) / NREP
    print(
        f"  Time for torch (torch.save):\t\t\t{tc:.3f} s ({(N * 8 / tc) / 2**30:.2f} GB/s)) ",
        end="",
    )
    buff.seek(0)
    c = buff.read()
    print(f"\tcr: {in_.size * in_.dtype.itemsize * 1.0 / len(c):5.1f}x")

    with open("serialize_torch.bin", "wb") as f:
        f.write(c)

    codec = blosc2.Codec.LZ4
    # print(f"Storing with {codec}")
    cparams = {"codec": codec, "clevel": 9}

    c = None
    ctic = time.time()
    for _i in range(NREP):
        c = blosc2.pack_tensor(in_, cparams=cparams)
    ctoc = time.time()
    tc = (ctoc - ctic) / NREP
    print(
        f"  Time for tensorflow (blosc2.pack_tensor):\t{tc:.3f} s ({(N * 8 / tc) / 2**30:.2f} GB/s)) ",
        end="",
    )
    print(f"\tcr: {in_.size * in_.dtype.itemsize * 1.0 / len(c):5.1f}x")

    with open("pack_tensorflow.bl2", "wb") as f:
        f.write(c)

    tt = torch.from_numpy(in_)
    c = None
    ctic = time.time()
    for _i in range(NREP):
        c = blosc2.pack_tensor(in_, cparams=cparams)
    ctoc = time.time()
    tc = (ctoc - ctic) / NREP
    print(
        f"  Time for torch (blosc2.pack_tensor):\t\t{tc:.3f} s ({(N * 8 / tc) / 2**30:.2f} GB/s)) ",
        end="",
    )
    print(f"\tcr: {in_.size * in_.dtype.itemsize * 1.0 / len(c):5.1f}x")

    with open("pack_torch.bl2", "wb") as f:
        f.write(c)

if True:
    with open("serialize_tensorflow.bin", "rb") as f:
        c = f.read()

    out = None
    dtic = time.time()
    for _i in range(NREP):
        out = tf.io.parse_tensor(c, out_type=in_.dtype)
    dtoc = time.time()
    td = (dtoc - dtic) / NREP
    print(
        f"  Time for tensorflow (tf.io.parse_tensor):\t{td:.3f} s ({(N * 8 / td) / 2**30:.2f} GB/s)) ",
    )

    with open("serialize_torch.bin", "rb") as f:
        buff = io.BytesIO(f.read())

    out = None
    dtic = time.time()
    for _i in range(NREP):
        buff.seek(0)
        out = torch.load(buff)
    dtoc = time.time()
    td = (dtoc - dtic) / NREP
    print(
        f"  Time for torch (torch.load):\t\t\t{td:.3f} s ({(N * 8 / td) / 2**30:.2f} GB/s)) ",
    )

    with open("pack_tensorflow.bl2", "rb") as f:
        c = f.read()

    out = None
    dtic = time.time()
    for _i in range(NREP):
        out = blosc2.unpack_tensor(c)
    dtoc = time.time()
    td = (dtoc - dtic) / NREP
    print(
        f"  Time for tensorflow (blosc2.unpack_tensor):\t{td:.3f} s ({(N * 8 / td) / 2**30:.2f} GB/s)) ",
    )
    assert np.array_equal(in_, out)

    with open("pack_torch.bl2", "rb") as f:
        c = f.read()

    out = None
    dtic = time.time()
    for _i in range(NREP):
        out = blosc2.unpack_tensor(c)

    dtoc = time.time()

    td = (dtoc - dtic) / NREP
    print(
        f"  Time for torch (blosc2.unpack_tensor):\t{td:.3f} s ({(N * 8 / td) / 2**30:.2f} GB/s)) ",
    )
    assert np.array_equal(in_, out)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################
from __future__ import annotations

import os
import pathlib
from collections import namedtuple
from collections.abc import Iterator, Mapping, MutableMapping
from dataclasses import asdict
from typing import Any, NamedTuple

import numpy as np
from msgpack import packb, unpackb

import blosc2
from blosc2 import SpecialValue, blosc2_ext


class vlmeta(MutableMapping, blosc2_ext.vlmeta):
    """
    Class providing access to user metadata on an :ref:`SChunk`.
    It is available via the `.vlmeta` property of an :ref:`SChunk`.
    """

    def __init__(self, schunk, urlpath, mode, mmap_mode, initial_mapping_size):
        self.urlpath = urlpath
        self.mode = mode
        self.mmap_mode = mmap_mode
        self.initial_mapping_size = initial_mapping_size
        super().__init__(schunk)

    def __setitem__(self, name, content):
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        # If name is a slice, assume that content is a dictionary and copy all the items
        if isinstance(name, slice):
            if name.start is None and name.stop is None:
                for k, v in content.items():
                    self.set_vlmeta(k, v)
                return
            raise NotImplementedError("Slicing is not supported, unless [:]")
        cparams = {"typesize": 1}
        content = packb(
            content,
            default=blosc2_ext.encode_tuple,
            strict_types=True,
            use_bin_type=True,
        )
        super().set_vlmeta(name, content, **cparams)

    def __getitem__(self, name):
        if isinstance(name, slice):
            if name.start is None and name.stop is None:
                # Return all the vlmetalayers
                return self.getall()
            raise NotImplementedError("Slicing is not supported, unless [:]")
        return unpackb(super().get_vlmeta(name), list_hook=blosc2_ext.decode_tuple)

    def __delitem__(self, name):
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        super().del_vlmeta(name)

    def __len__(self):
        return super().nvlmetalayers()

    def __iter__(self):
        keys = super().get_names()
        yield from keys

    def getall(self):
        """
        Return all the variable length metalayers as a dictionary

        """
        return super().to_dict()

    def __repr__(self):
        return repr(self.getall())

    def __str__(self):
        return str(self.getall())


class Meta(Mapping):
    """
    Class providing access to fixed-length metadata on an :ref:`SChunk`.
    It is available via the `.meta` property of an :ref:`SChunk`.
    """

    def get(self, key: str, default: Any = None) -> Any:
        """Return the value for `key` if `key` is in the dictionary, else return `default`.
        If `default` is not given, it defaults to ``None``."""
        return self.get(key, default)

    def __init__(self, schunk):
        self.schunk = schunk

    def __contains__(self, key: str) -> bool:
        """Check if the `key` metalayer exists or not."""
        return blosc2_ext.meta__contains__(self.schunk, key)

    def __delitem__(self, key: str) -> None:
        raise NotImplementedError("Cannot remove a metalayer")

    def __setitem__(self, key: str, value: bytes) -> None:
        """Update the `key` metalayer with `value`.

        Parameters
        ----------
        key: str
            The name of the metalayer to update.
        value: bytes
            The buffer containing the new content for the metalayer.

            ..warning: Note that the *length* of the metalayer cannot change,
            otherwise an exception will be raised.
        """
        value = packb(value, default=blosc2_ext.encode_tuple, strict_types=True, use_bin_type=True)
        blosc2_ext.meta__setitem__(self.schunk, key, value)

    def __getitem__(self, item: str | slice) -> bytes | dict[str, bytes]:
        """Return the specified metalayer.

        Parameters
        ----------
        item: str or slice
            The name of the metalayer to return.  If a slice is passed,
            and start and stop are None ([:]), all the metalayers are returned;
            else, a NotImplementedError is raised.

        Returns
        -------
        bytes or dict
            The buffer containing the metalayer information. If a slice is passed,
            a dictionary with all the metalayers is returned.
        """
        if isinstance(item, slice):
            if item.start is None and item.stop is None:
                return self.getall()
            raise NotImplementedError("Slicing is not supported, unless [:]")
        if self.__contains__(item):
            return unpackb(
                blosc2_ext.meta__getitem__(self.schunk, item),
                list_hook=blosc2_ext.decode_tuple,
            )
        else:
            raise KeyError(f"{item} not found")

    def keys(self) -> list[str]:
        """Return the metalayers keys."""
        return blosc2_ext.meta_keys(self.schunk)

    def values(self):
        raise NotImplementedError("Values can not be accessed")

    def items(self):
        raise NotImplementedError("Items can not be accessed")

    def __iter__(self) -> Iterator[str]:
        """Iter over the keys of the metalayers."""
        return iter(self.keys())

    def __len__(self) -> int:
        """Return the number of metalayers."""
        return blosc2_ext.meta__len__(self.schunk)

    def getall(self):
        """
        Return all the variable length metalayers as a dictionary

        """
        return {key: self[key] for key in self.keys()}

    def __repr__(self):
        return repr(self.getall())

    def __str__(self):
        return str(self.getall())


class SChunk(blosc2_ext.SChunk):
    def __init__(  # noqa: C901
        self,
        chunksize: int | None = None,
        data: object = None,
        **kwargs: dict | blosc2.CParams | blosc2.Storage | blosc2.DParams,
    ) -> None:
        """Create a new super-chunk, or open an existing one.

        Parameters
        ----------
        chunksize: int, optional
            The size, in bytes, of the chunks in the super-chunk. If not provided,
            it is set automatically to a reasonable value.

        data: bytes-like object, optional
            The data to be split into different chunks of size :paramref:`chunksize`.
            If None, the Schunk instance will be empty initially.

        kwargs: dict, optional
            Storage parameters. The default values are in :class:`blosc2.Storage`.
            Supported keyword arguments:
                storage: :class:`blosc2.Storage` or dict
                    All the storage parameters that you want to use as
                    a :class:`blosc2.Storage` or dict instance.
                cparams: :class:`blosc2.CParams` or dict
                    All the compression parameters that you want to use as
                    a :class:`blosc2.CParams` or dict instance.
                dparams: :class:`blosc2.DParams` or dict
                    All the decompression parameters that you want to use as
                    a :class:`blosc2.DParams` or dict instance.
                others: Any
                    If `storage` is not passed, all the parameters of a :class:`blosc2.Storage`
                    can be passed as keyword arguments.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> import os.path
        >>> import shutil
        >>> import tempfile
        >>> cparams = blosc2.CParams()
        >>> dparams = blosc2.DParams()
        >>> storage = blosc2.Storage(contiguous=True)
        >>> schunk = blosc2.SChunk(cparams=cparams, dparams=dparams, storage=storage)

        In the following, we will write and read a super-chunk to and from disk
        via memory-mapped files.

        >>> a = np.arange(3, dtype=np.int64)
        >>> chunksize = a.size * a.itemsize
        >>> n_chunks = 2
        >>> tmpdirname = tempfile.mkdtemp()
        >>> urlpath = os.path.join(tmpdirname, 'schunk.b2frame')

        Optional: we intend to write 2 chunks of 24 bytes each, and we expect
        the compressed size to be smaller than the original size. Therefore, we
        generously set the initial size of the mapping to 48 bytes
        effectively avoiding remappings.

        >>> initial_mapping_size = chunksize * n_chunks
        >>> schunk_mmap = blosc2.SChunk(
        ...     chunksize=chunksize,
        ...     mmap_mode="w+",
        ...     initial_mapping_size=initial_mapping_size,
        ...     urlpath=urlpath,
        ... )
        >>> schunk_mmap.append_data(a)
        1
        >>> schunk_mmap.append_data(a * 2)
        2

        Optional: explicitly close the file and free the mapping.

        >>> del schunk_mmap

        Reading the data back again via memory-mapped files:

        >>> schunk_mmap = blosc2.open(urlpath, mmap_mode="r")
        >>> np.frombuffer(schunk_mmap.decompress_chunk(0), dtype=np.int64).tolist()
        [0, 1, 2]
        >>> np.frombuffer(schunk_mmap.decompress_chunk(1), dtype=np.int64).tolist()
        [0, 2, 4]
        >>> shutil.rmtree(tmpdirname)
        """
        # Check only allowed kwarg are passed
        allowed_kwargs = [
            "urlpath",
            "contiguous",
            "cparams",
            "dparams",
            "_schunk",
            "meta",
            "mode",
            "mmap_mode",
            "initial_mapping_size",
            "_is_view",
            "storage",
        ]
        for kwarg in kwargs:
            if kwarg not in allowed_kwargs:
                raise ValueError(f"{kwarg} is not supported as keyword argument")
        if kwargs.get("storage") is not None:
            if any(key in list(blosc2.Storage.__annotations__) for key in kwargs):
                raise AttributeError(
                    "Cannot pass both `storage` and other kwargs already included in Storage"
                )
            storage = kwargs.get("storage")
            if isinstance(storage, blosc2.Storage):
                kwargs = {**kwargs, **asdict(storage)}
            else:
                kwargs = {**kwargs, **storage}

        if isinstance(kwargs.get("cparams"), blosc2.CParams):
            kwargs["cparams"] = asdict(kwargs.get("cparams"))

        if isinstance(kwargs.get("dparams"), blosc2.DParams):
            kwargs["dparams"] = asdict(kwargs.get("dparams"))

        urlpath = kwargs.get("urlpath")
        if "contiguous" not in kwargs:
            # Make contiguous true for disk, else sparse (for in-memory performance)
            kwargs["contiguous"] = urlpath is not None

        # This a private param to get an SChunk from a blosc2_schunk*
        sc = kwargs.pop("_schunk", None)

        # If not passed, set a sensible typesize
        itemsize = data.itemsize if data is not None and hasattr(data, "itemsize") else 1
        if "cparams" in kwargs:
            if "typesize" not in kwargs["cparams"]:
                cparams = kwargs.pop("cparams").copy()
                cparams["typesize"] = itemsize
                kwargs["cparams"] = cparams
        else:
            kwargs["cparams"] = {"typesize": itemsize}

        # chunksize handling
        if chunksize is None:
            chunksize = 2**24
            if data is not None:
                if hasattr(data, "itemsize"):
                    chunksize = data.size * data.itemsize
                    # Make that a multiple of typesize
                    chunksize = chunksize // data.itemsize * data.itemsize
                else:
                    chunksize = len(data)
            # Use a cap of 256 MB (modern boxes should all have this RAM available)
            if chunksize > 2**28:
                chunksize = 2**28

        super().__init__(_schunk=sc, chunksize=chunksize, data=data, **kwargs)
        self._vlmeta = vlmeta(
            super().c_schunk, self.urlpath, self.mode, self.mmap_mode, self.initial_mapping_size
        )
        self._cparams = super().get_cparams()
        self._dparams = super().get_dparams()

    @property
    def cparams(self) -> blosc2.CParams:
        """
        :class:`blosc2.CParams` instance with the compression parameters.
        """
        return self._cparams

    @cparams.setter
    def cparams(self, value: blosc2.CParams) -> None:
        super().update_cparams(value)
        self._cparams = super().get_cparams()

    @property
    def dparams(self) -> blosc2.DParams:
        """
        :class:`blosc2.DParams` instance with the decompression parameters.
        """
        return self._dparams

    @dparams.setter
    def dparams(self, value: blosc2.DParams) -> None:
        super().update_dparams(value)
        self._dparams = super().get_dparams()

    @property
    def meta(self) -> Meta:
        """
        Access to the fixed-length metadata of the `SChunk`.
        """
        return Meta(self)

    @property
    def vlmeta(self) -> vlmeta:
        """
        Access to the variable-length metadata of the `SChunk`.
        """
        return self._vlmeta

    @property
    def chunkshape(self) -> int:
        """
        Number of elements per chunk.
        """
        return self.chunksize // self.typesize

    @property
    def chunksize(self) -> int:
        """
        Number of bytes in each chunk.
        """
        return super().chunksize

    @property
    def blocksize(self) -> int:
        """The block size (in bytes)."""
        return super().blocksize

    @property
    def nchunks(self) -> int:
        """The number of chunks."""
        return super().nchunks

    @property
    def cratio(self) -> float:
        """
        Compression ratio.
        """
        if self.cbytes == 0:
            return 0.0
        return self.nbytes / self.cbytes

    @property
    def nbytes(self) -> int:
        """
        Amount of uncompressed data bytes.
        """
        return super().nbytes

    @property
    def cbytes(self) -> int:
        """
        Amount of compressed data bytes (data size + chunk headers size).
        """
        return super().cbytes

    @property
    def typesize(self) -> int:
        """
        Type size of the `SChunk`.
        """
        return super().typesize

    @property
    def urlpath(self) -> str:
        """
        Path where the `SChunk` is stored.
        """
        return super().urlpath

    @property
    def contiguous(self) -> bool:
        """
        Whether the `SChunk` is stored contiguously or sparsely.
        """
        return super().contiguous

    def append_data(self, data: object) -> int:
        """Append a data buffer to the SChunk.

        The data buffer must be of size `chunksize` specified in
        :func:`SChunk.__init__ <blosc2.schunk.SChunk.__init__>`.

        Parameters
        ----------
        data: bytes-like object
            The data to be compressed and added as a chunk.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If the :paramref:`data` could not be appended.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> schunk = blosc2.SChunk(chunksize=200*1000*4)
        >>> data = np.arange(200 * 1000, dtype='int32')
        >>> schunk.append_data(data)
        1
        """
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().append_data(data)

    def fill_special(
        self,
        nitems: int,
        special_value: blosc2.SpecialValue,
        value: bytes | int | float | bool | None = None,
    ) -> int:
        """Fill the SChunk with a special value. The SChunk must be empty.

        Parameters
        ----------
        nitems: int
            The number of items to fill with the special value.
        special_value: SpecialValue
            The special value to be used for filling the SChunk.
        value: bytes, int, float, bool (optional)
            The value to fill the SChunk. This parameter is only supported if
            :paramref:`special_value` is ``blosc2.SpecialValue.VALUE``.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If the SChunk could not be filled with the special value.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> import time
        >>> nitems = 100_000_000
        >>> dtype = np.dtype(np.float64)
        >>> # Measure the time to create SChunk from a NumPy array
        >>> t0 = time.time()
        >>> data = np.full(nitems, np.pi, dtype)
        >>> cparams = blosc2.CParams(typesize=dtype.itemsize)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> t = (time.time() - t0) * 1000.
        >>> f"Time creating a schunk with a numpy array: {t:10.3f} ms"
        Time creating a schunk with a numpy array:    710.273 ms
        >>> # Measure the time to create SChunk using fill_special
        >>> t0 = time.time()
        >>> cparams = blosc2.CParams(typesize=dtype.itemsize)
        >>> schunk = blosc2.SChunk(cparams=cparams)
        >>> schunk.fill_special(nitems, blosc2.SpecialValue.VALUE, np.pi)
        >>> t = (time.time() - t0) * 1000.
        >>> f"Time passing directly the value to `fill_special`: {t:10.3f} ms"
        Time passing directly the value to `fill_special`:      2.109 ms
        """
        if not isinstance(special_value, SpecialValue) or special_value == SpecialValue.NOT_SPECIAL:
            raise TypeError("special_value must be a SpecialValue instance other than NOT_SPECIAL")
        if special_value == SpecialValue.VALUE and value is None:
            raise ValueError("value cannot be None when special_value is VALUE")

        nchunks = super().fill_special(nitems, special_value.value, value)
        if nchunks < 0:
            raise RuntimeError("Unable to fill with special values")
        return nchunks

    def decompress_chunk(self, nchunk: int, dst: object = None) -> str | bytes:
        """Decompress the chunk given by its index :paramref:`nchunk`.

        Parameters
        ----------
        nchunk: int
            The index of the chunk that will be decompressed.
        dst: NumPy object or bytearray
            The destination NumPy object or bytearray to fill, the length
            of which must be greater than 0. The user must ensure
            that it has enough capacity to host the decompressed
            chunk. Default is None, meaning that a new bytes object
            is created, filled and returned.

        Returns
        -------
        out: str or bytes
            The decompressed chunk as a Python str or bytes object if
            :paramref:`dst` is `None`. Otherwise, it returns `None` because the
            result will already be in :paramref:`dst`.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> cparams = blosc2.CParams(typesize=1)
        >>> schunk = blosc2.SChunk(cparams=cparams)
        >>> buffer = b"wermqeoir23"
        >>> schunk.append_data(buffer)
        1
        >>> schunk.decompress_chunk(0)
        b'wermqeoir23'
        >>> # Construct a mutable bytearray object
        >>> bytes_obj = bytearray(len(buffer))
        >>> schunk.decompress_chunk(0, dst=bytes_obj)
        >>> bytes_obj == buffer
        True
        """
        return super().decompress_chunk(nchunk, dst)

    def get_chunk(self, nchunk: int) -> bytes:
        """Return the compressed chunk that is in the SChunk.

        Parameters
        ----------
        nchunk: int
            The index of the chunk that will be returned.

        Returns
        -------
        out: bytes object
            The compressed chunk.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> # Create an SChunk with 3 chunks
        >>> nchunks = 3
        >>> data = np.arange(200 * 1000 * nchunks, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> # Retrieve the first chunk (index 0)
        >>> chunk = schunk.get_chunk(0)
        >>> # Check the type and length of the compressed chunk
        >>> type(chunk)
        <class 'bytes'>
        >>> len(chunk)
        10552
        """
        return super().get_chunk(nchunk)

    def delete_chunk(self, nchunk: int) -> int:
        """Delete the specified chunk from the SChunk.

        Parameters
        ----------
        nchunk: int
            The index of the chunk that will be removed.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> # Create an SChunk with 3 chunks
        >>> nchunks = 3
        >>> data = np.arange(200 * 1000 * nchunks, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(chunksize=200 * 1000 * 4, data=data, cparams=cparams)
        >>> # Check the number of chunks before deletion
        >>> schunk.nchunks
        3
        >>>  # Delete the second chunk (index 1)
        >>> schunk.delete_chunk(1)
        >>>  # Check the number of chunks after deletion
        >>> schunk.nchunks
        2
        """
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().delete_chunk(nchunk)

    def insert_chunk(self, nchunk: int, chunk: bytes) -> int:
        """Insert an already compressed chunk into the SChunk.

        Parameters
        ----------
        nchunk: int
            The index at which the chunk will be inserted.
        chunk: bytes object
            The compressed chunk.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> # Create an SChunk with 2 chunks
        >>> data = np.arange(400 * 1000, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(chunksize=200*1000*4, data=data, cparams=cparams)
        >>> # Get a compressed chunk from the SChunk
        >>> chunk = schunk.get_chunk(0)
        >>> # Insert a chunk in the second position (index 1)"
        >>> schunk.insert_chunk(1, chunk)
        >>> # Verify the total number of chunks after insertion
        >>> schunk.nchunks
        3
        """
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().insert_chunk(nchunk, chunk)

    def insert_data(self, nchunk: int, data: object, copy: bool) -> int:
        """Insert the data in the specified position in the SChunk.

        Parameters
        ----------
        nchunk: int
            The index at which the chunk will be inserted.
        data: bytes object
            The data that will be compressed and inserted as a chunk.
        copy: bool
            Whether to make an internal copy of the chunk to insert it or not.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> # Create an SChunk with 2 chunks
        >>> data = np.arange(400 * 1000, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(chunksize=200*1000*4, data=data, cparams=cparams)
        >>> # Create a new array to insert into the second chunk of the SChunk
        >>> new_data = np.arange(200 * 1000, dtype=np.int32)
        >>> # Insert the new data at position 1, compressing it
        >>> schunk.insert_data(1, new_data, copy=True)
        >>> # Verify the total number of chunks after insertion
        >>> schunk.nchunks
        3
        """
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().insert_data(nchunk, data, copy)

    def update_chunk(self, nchunk: int, chunk: bytes) -> int:
        """Update an existing chunk in the SChunk.

        Parameters
        ----------
        nchunk: int
            The index of the chunk to be updated.
        chunk: bytes object
            The new compressed chunk that will replace the old chunk's content.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> nchunks = 5
        >>> chunk_size = 200 * 1000 * 4
        >>> data = np.arange(nchunks * chunk_size // 4, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(chunksize=chunk_size, data=data, cparams=cparams)
        >>> f"Initial number of chunks: {schunk.nchunks}"
        Initial number of chunks: 5
        >>> c_index = 1
        >>> new_data = np.full(chunk_size // 4, fill_value=c_index, dtype=np.int32).tobytes()
        >>> compressed_data = blosc2.compress2(new_data, typesize=4)
        >>> # Update the 2nd chunk (index 1) with new data
        >>> nchunks = schunk.update_chunk(c_index, compressed_data)
        >>> f"Number of chunks after update: {nchunks}"
        Number of chunks after update: 5
        """
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().update_chunk(nchunk, chunk)

    def update_data(self, nchunk: int, data: object, copy: bool) -> int:
        """Update the chunk in the specified position with the given data.

        Parameters
        ----------
        nchunk: int
            The index of the chunk to be updated.
        data: bytes object
            The data to be compressed and will replace the old chunk.
        copy: bool
             Whether to make an internal copy of the chunk before updating it.

        Returns
        -------
        out: int
            The number of chunks in the SChunk.

        Raises
        ------
        RunTimeError
            If a problem is detected.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> nchunks = 4
        >>> chunk_size = 200 * 1000 * 4
        >>> data = np.arange(nchunks * chunk_size // 4, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(chunksize=chunk_size, data=data, cparams=cparams)
        >>> f"Initial number of chunks: {schunk.nchunks}"
        Initial number of chunks: 4
        >>> c_index = 1 # Update the 2nd chunk (index 1)
        >>> new_data = np.full(chunk_size // 4, fill_value=c_index, dtype=np.int32).tobytes()
        >>> nchunks = schunk.update_data(c_index, new_data, copy=True)
        >>> f"Number of chunks after update: {schunk.nchunks}"
        Number of chunks after update: 4
        """
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().update_data(nchunk, data, copy)

    def get_slice(self, start: int = 0, stop: int | None = None, out: object = None) -> str | bytes | None:
        """Get a slice from :paramref:`start` to :paramref:`stop`.

        Parameters
        ----------
        start: int
            The starting index of the slice. Default is 0.
        stop: int
            The ending index of the slice (exclusive).
            Default is until the SChunk ends.
        out: bytes-like object or bytearray
            The target object (supporting the
            `Buffer Protocol <https://docs.python.org/3/c-api/buffer.html>`_) to fill.
            Verify that the buffer has enough space for the decompressed data.
            If `None` is provided, a new bytes object will be created, filled,
            and returned.

        Returns
        -------
        out: str or bytes or None
            The decompressed slice a Python str or bytes object if
            :paramref:`out` is `None`. Otherwise, it returns `None` since the result
            will already be in :paramref:`out`.

        Raises
        ------
        ValueError
            If the size to get is negative.
            If there is not enough space in :paramref:`out`.
            If :paramref:`start` is greater or equal to the number of items in the SChunk.
        RunTimeError
            If a problem is detected.

        See Also
        --------
        :func:`__getitem__`

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> nchunks = 4
        >>> chunk_size = 200 * 1000 * 4
        >>> data = np.arange(nchunks * chunk_size // 4, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> # Define the slice parameters
        >>> start_index = 200 * 1000
        >>> stop_index = 2 * 200 * 1000
        >>> # Prepare an output buffer
        >>> slice_size = stop_index - start_index
        >>> out_buffer = bytearray(slice_size * 4)  # Ensure the buffer is large enough
        >>> result = schunk.get_slice(start=start_index, stop=stop_index, out=out_buffer)
        >>> # Convert bytearray to NumPy array for easier inspection
        >>> slice_array = np.frombuffer(out_buffer, dtype=np.int32)
        >>> f"Slice data: {slice_array[:10]} ..."  # Print the first 10 elements
        Slice data: [200000 200001 200002 200003 200004 200005 200006 200007 200008 200009] ...
        """
        return super().get_slice(start, stop, out)

    def __len__(self) -> int:
        """
        Return the number of items in the SChunk.
        """
        return self.nbytes // self.typesize

    def __getitem__(self, item: int | slice) -> str | bytes:
        """Get a slice from the SChunk.

        Parameters
        ----------
        item: int or slice
            The index or slice for the data. Note that the step parameter is not honored.

        Returns
        -------
        out: str or bytes
            The decompressed slice as a Python str or bytes object.

        Raises
        ------
        ValueError
            If the size to get is negative.
            If :paramref:`item`.start is greater than or equal to the number of
            items in the SChunk.
        RunTimeError
            If a problem is detected.
        IndexError
            If `step` is not 1.

        See Also
        --------
        :func:`get_slice`

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> nchunks = 4
        >>> chunk_size = 200 * 1000 * 4
        >>> data = np.arange(nchunks * chunk_size // 4, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(chunksize=chunk_size, data=data, cparams=cparams)
        >>> # Use __getitem__ to retrieve the same slice of data from the SChunk
        >>> res = schunk[150:155]
        >>> f"Slice data: {np.frombuffer(res, dtype=np.int32)}"
        Slice data: [150 151 152 153 154]
        """
        if isinstance(item, int):
            if item == -1:
                return self.get_slice(item)
            return self.get_slice(item, item + 1)
        if item.step is not None and item.step != 1:
            raise IndexError("`step` must be 1")
        return self.get_slice(item.start, item.stop)

    def __setitem__(self, key: int | slice, value: object) -> None:
        """Set slice to :paramref:`value`.

        Parameters
        ----------
        key: int or slice
            The index of the slice to update. Note that step parameter is not honored.
        value: bytes-like object
            An object supporting the
            `Buffer Protocol <https://docs.python.org/3/c-api/buffer.html>`_ used to
            fill the slice.

        Returns
        -------
        out: None

        Raises
        ------
        ValueError
            If the object cannot be modified.
            If the size to get is negative.
            If there is not enough space in :paramref:`value` to update the slice.
            If :paramref:`start` is greater than the number of items in the SChunk.
        RunTimeError
            If a problem is detected.
        IndexError
            If `step` is not 1.

        Notes
        -----
        This method can also be used to append new data if :paramref:`key`.stop
        is greater than the number of items in the SChunk.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> nchunks = 4
        >>> chunk_size = 200 * 1000 * 4
        >>> data = np.arange(nchunks * chunk_size // 4, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> # Create a new array of values to update the slice (values from 1000 to 1999 multiplied by 2)
        >>> start_ = 1000
        >>> stop = 2000
        >>> new_values = np.arange(start_, stop, dtype=np.int32) * 2
        >>> schunk[start_:stop] = new_values
        >>> # Retrieve the updated slice using the slicing syntax
        >>> retrieved_slice = np.frombuffer(schunk[start_:stop], dtype=np.int32)
        >>> f"First 10 values of the updated slice: {retrieved_slice[:10]}"
        >>> f"Last 10 values of the updated slice: {retrieved_slice[-10:]}"
        First 10 values of the updated slice: [2000 2002 2004 2006 2008 2010 2012 2014 2016 2018]
        Last 10 values of the updated slice: [3980 3982 3984 3986 3988 3990 3992 3994 3996 3998]
        """
        if key.step is not None and key.step != 1:
            raise IndexError("`step` must be 1")
        blosc2_ext.check_access_mode(self.urlpath, self.mode)
        return super().set_slice(start=key.start, stop=key.stop, value=value)

    def to_cframe(self) -> bytes:
        """Get a bytes object containing the serialized :ref:`SChunk` instance.

        Returns
        -------
        out: bytes
            The buffer containing the serialized :ref:`SChunk` instance.

        See Also
        --------
        :func:`~blosc2.schunk_from_cframe`

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> nchunks = 4
        >>> chunk_size = 200 * 1000 * 4
        >>> data = np.arange(nchunks * chunk_size // 4, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> # Serialize the SChunk instance to a bytes object
        >>> serialized_schunk = schunk.to_cframe()
        >>> f"Serialized SChunk length: {len(serialized_schunk)} bytes"
        Serialized SChunk length: 14129 bytes
        >>> # Create a new SChunk from the serialized data
        >>> deserialized_schunk = blosc2.schunk_from_cframe(serialized_schunk)
        >>> start = 500
        >>> stop = 505
        >>> sl_bytes = deserialized_schunk[start:stop]
        >>> sl = np.frombuffer(sl_bytes, dtype=np.int32)
        >>> res = data[start:stop]
        >>> f"Original slice: {res}"
        Original slice: [500 501 502 503 504]
        >>> f"Deserialized slice: {sl}"
        Deserialized slice: [500 501 502 503 504]
        """
        return super().to_cframe()

    def iterchunks(self, dtype: np.dtype) -> Iterator[np.ndarray]:
        """
        Iterate over the :paramref:`self` chunks of the SChunk.

        Parameters
        ----------
        dtype: np.dtype
            The data type to use for the decompressed chunks.

        Yields
        ------
        chunk: NumPy ndarray
           The decompressed chunk.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> # Create sample data and an SChunk
        >>> data = np.arange(400 * 1000, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> # Iterate over chunks using the iterchunks method
        >>> for chunk in schunk.iterchunks(dtype=np.int32):
        >>>     f"Chunk shape: {chunk.shape} "
        >>>     f"First 5 elements of chunk: {chunk[:5]}"
        Chunk shape: (400000,)
        First 5 elements of chunk: [0 1 2 3 4]
        """
        out = np.empty(self.chunkshape, dtype)
        for i in range(0, len(self), self.chunkshape):
            self.get_slice(i, i + self.chunkshape, out)
            yield out

    def iterchunks_info(
        self,
    ) -> Iterator[
        NamedTuple(
            "info",
            nchunk=int,
            cratio=float,
            special=blosc2.SpecialValue,
            repeated_value=bytes | None,
            lazychunk=bytes,
        )
    ]:
        """
        Iterate over the chunks of the SChunk, providing info on index and special values.

        Yields
        ------
        info: namedtuple
            A namedtuple with the following fields:

                nchunk: int
                    The index of the chunk.
                cratio: float
                    The compression ratio of the chunk.
                special: :class:`~blosc2.SpecialValue`
                    The special value enum of the chunk; if 0, the chunk is not special.
                repeated_value: bytes or None
                    The repeated value for the chunk; if not SpecialValue.VALUE, it is None.
                lazychunk: bytes
                    A buffer with the complete lazy chunk.

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> # Create sample data and an SChunk
        >>> data = np.arange(400 * 1000, dtype=np.int32)
        >>> cparams = blosc2.CParams(typesize=4)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams)
        >>> # Iterate over chunks and print detailed information
        >>> for chunk_info in schunk.iterchunks_info():
        >>>     f"Chunk index: {chunk_info.nchunk}"
        >>>     f"Compression ratio: {chunk_info.cratio:.2f}"
        >>>     f"Special value: {chunk_info.special.name}"
        >>>     f"Repeated value: {chunk_info.repeated_value[:10] if chunk_info.repeated_value else None}"
        Chunk index: 0
        Compression ratio: 223.56
        Special value: NOT_SPECIAL
        Repeated value: None
        """
        ChunkInfo = namedtuple("ChunkInfo", ["nchunk", "cratio", "special", "repeated_value", "lazychunk"])
        for nchunk in range(self.nchunks):
            lazychunk = self.get_lazychunk(nchunk)
            # Blosc2 flags are encoded at the end of the header
            # (see https://github.com/Blosc/c-blosc2/blob/main/README_CHUNK_FORMAT.rst)
            is_special = (lazychunk[31] & 0x70) >> 4
            special = SpecialValue(is_special)
            # The special value is encoded at the end of the header
            repeated_value = lazychunk[32:] if special == SpecialValue.VALUE else None
            # Compression ratio (nbytes and cbytes are little-endian)
            cratio = (
                np.frombuffer(lazychunk[4:8], dtype="<i4")[0]
                / np.frombuffer(lazychunk[12:16], dtype="<i4")[0]
            )
            yield ChunkInfo(nchunk, cratio, special, repeated_value, lazychunk)

    def postfilter(self, input_dtype: np.dtype, output_dtype: np.dtype = None) -> None:
        """Decorator to set a function as a postfilter.

        The postfilter function will be executed each time after decompressing
        blocks of data. It will receive three parameters:

        * the input `ndarray` to be read from
        * the output `ndarray` to be filled out
        * the offset inside the `SChunk` instance where the corresponding block begins (see example below).

        Parameters
        ----------
        input_dtype: np.dtype
            Data type of the input that will receive the postfilter function.
        output_dtype: np.dtype
            Data type of the output that will receive and fill the postfilter function.
            If None (default) it will be set to :paramref:`input_dtype`.

        Returns
        -------
        out: None

        Notes
        -----
        * `nthreads` must be 1 when decompressing.

        * The :paramref:`input_dtype` itemsize must be the same as the
          :paramref:`output_dtype` itemsize.

        See Also
        --------
        :meth:`remove_postfilter`
        :meth:`prefilter`

        Examples
        --------
        .. code-block:: python

            # Create SChunk
            input_dtype = np.dtype(np.int64)
            cparams = blosc2.CParams(typesize=input_dtype.itemsize)
            dparams = blosc2.DParams(nthreads=1)
            schunk = blosc2.SChunk(
                chunksize=20_000 * input_dtype.itemsize, cparams=cparams, dparams=dparams
            )


            # Create postfilter and associate it to the schunk
            @schunk.postfilter(input_dtype)
            def postfilter(input, output, offset):
                output[:] = offset + np.arange(input.size)
        """

        def initialize(func):
            super(SChunk, self)._set_postfilter(func, input_dtype, output_dtype)

            def exec_func(*args):
                func(*args)

            return exec_func

        return initialize

    def remove_postfilter(self, func_name: str, _new_ctx: bool = True) -> None:
        """Remove the postfilter from the `SChunk` instance.

        Parameters
        ----------
        func_name: str
            The name of the postfilter function to remove.

        Returns
        -------
        out: None

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> dtype = np.dtype(np.int32)
        >>> cparams = blosc2.CParams(typesize=dtype.itemsize)
        >>> dparams = blosc2.DParams(nthreads=1)
        >>> data = np.arange(500, dtype=np.int32)
        >>> schunk = blosc2.SChunk(data=data, cparams=cparams, dparams=dparams)
        >>> # Define the postfilter function
        >>> @schunk.postfilter(dtype)
        >>> def postfilter(input, output, offset):
        >>>     output[:] = input + offset + np.arange(input.size)
        >>> out = np.empty(data.size, dtype=dtype)
        >>> schunk.get_slice(out=out)
        >>> f"Data slice with postfilter applied (first 8 elements): {out[:8]}"
        Data slice with postfilter applied (first 8 elements): [ 0  2  4  6  8 10 12 14]
        >>> schunk.remove_postfilter('postfilter')
        >>> retrieved_data = np.empty(data.size, dtype=dtype)
        >>> schunk.get_slice(out=retrieved_data)
        >>> f"Original data (first 8 elements): {data[:8]}"
        Original data (first 8 elements): [0 1 2 3 4 5 6 7]
        """
        return super().remove_postfilter(func_name)

    def filler(self, inputs_tuple: tuple[tuple], schunk_dtype: np.dtype, nelem: int | None = None) -> None:
        """Decorator to set a filler function.

        This function will fill :paramref:`self` according to :paramref:`nelem`.
        It will receive three parameters: a tuple with the inputs as `ndarrays`
        from which to read, the `ndarray` to fill :paramref:`self` and the
        offset inside the `SChunk` instance where the corresponding block
        begins (see example below).

        Parameters
        ----------
        inputs_tuple: tuple of tuples
            Tuple containing a tuple for each argument that the function will receive, along with their
            corresponding np.dtype.
            Supported operand types are :ref:`SChunk`, `ndarray` and
            Python scalars.
        schunk_dtype: np.dtype
            The data type to use to fill :paramref:`self`.
        nelem: int
            Number of elements to append to :paramref:`self`. If None (default) it
            will be the number of elements from the operands.

        Returns
        -------
        out: None

        Notes
        -----
        * Compression `nthreads` must be 1 when using this.
        * This does not need to be removed from the created `SChunk` instance.

        See Also
        --------
        :meth:`prefilter`

        Examples
        --------
        .. code-block:: python

            # Set the compression and decompression parameters
            schunk_dtype = np.dtype(np.float64)
            cparams = blosc2.CParams(typesize=schunk_dtype.itemsize, nthreads=1)
            # Create empty SChunk
            schunk = blosc2.SChunk(chunksize=20_000 * schunk_dtype.itemsize, cparams=cparams)

            # Create operands
            op_dtype = np.dtype(np.int32)
            data = np.full(20_000 * 3, 12, dtype=op_dtype)
            schunk_op = blosc2.SChunk(chunksize=20_000 * op_dtype.itemsize, data=data)


            # Create filler
            @schunk.filler(((schunk_op, op_dtype), (np.e, np.float32)), schunk_dtype)
            def filler(inputs_tuple, output, offset):
                output[:] = inputs_tuple[0] - inputs_tuple[1]

        """

        def initialize(func):
            if self.nbytes != 0:
                raise ValueError("Cannot apply a filler to a non empty SChunk")
            nelem_ = blosc2_ext.nelem_from_inputs(inputs_tuple, nelem)
            super(SChunk, self)._set_filler(func, id(inputs_tuple), schunk_dtype)
            chunksize = self.chunksize
            written_nbytes = 0
            nbytes = nelem_ * self.typesize
            while written_nbytes < nbytes:
                chunk = np.zeros(chunksize // self.typesize, dtype=schunk_dtype)
                self.append_data(chunk)
                written_nbytes += chunksize
                if (nbytes - written_nbytes) < self.chunksize:
                    chunksize = nbytes - written_nbytes
            self.remove_prefilter(func.__name__)

            def exec_func(*args):
                func(*args)

            return exec_func

        return initialize

    def prefilter(self, input_dtype: np.dtype, output_dtype: np.dtype = None) -> None:
        """Decorator to set a function as a prefilter.

        This function will be executed each time before compressing the data.
        It will receive three parameters:

        * The actual data as a `ndarray` from which to read,
        * The `ndarray` to be filled,
        * The offset inside the `SChunk` instance where the corresponding block begins (see example below).

        Parameters
        ----------
        input_dtype: np.dtype
            Data type of the input that will be processed the prefilter function.
        output_dtype: np.dtype, optional
            Data type of the output that will be filled by the prefilter function.
            If None (default), it will be the same as :paramref:`input_dtype`.

        Returns
        -------
        out: None

        Notes
        -----
        * `nthreads` must be 1 when compressing.

        * The :paramref:`input_dtype` itemsize must be the same as the
          :paramref:`output_dtype` itemsize.

        See Also
        --------
        :meth:`remove_prefilter`
        :meth:`postfilter`
        :meth:`filler`

        Examples
        --------
        .. code-block:: python

            # Set the compression and decompression parameters
            input_dtype = np.dtype(np.int32)
            output_dtype = np.dtype(np.float32)
            cparams = blosc2.CParams(typesize=output_dtype.itemsize, nthreads=1)
            # Create schunk
            schunk = blosc2.SChunk(chunksize=200 * 1000 * input_dtype.itemsize, cparams=cparams)


            # Set prefilter with decorator
            @schunk.prefilter(input_dtype, output_dtype)
            def prefilter(input, output, offset):
                output[:] = input - np.pi
        """

        def initialize(func):
            super(SChunk, self)._set_prefilter(func, input_dtype, output_dtype)

            def exec_func(*args):
                func(*args)

            return exec_func

        return initialize

    def remove_prefilter(self, func_name: str, _new_ctx: bool = True) -> None:
        """Remove the prefilter from the `SChunk` instance.

        Parameters
        ----------
        func_name: str
            Name of the prefilter function.

        Returns
        -------
        out: None

        Examples
        --------
        >>> import blosc2
        >>> import numpy as np
        >>> dtype = np.dtype(np.int32)
        >>> cparams = blosc2.CParams(typesize=dtype.itemsize, nthreads=1)
        >>> data = np.arange(1000, dtype=np.int32)
        >>> output_dtype = np.float32
        >>> schunk = blosc2.SChunk(cparams=cparams)
        >>> # Define the prefilter function
        >>> @schunk.prefilter(dtype, output_dtype)
        >>> def prefilter(input, output, offset):
        >>>     output[:] = input - np.pi
        >>> schunk[:1000] = data
        >>> # Retrieve and convert compressed data with the prefilter to a NumPy array.
        >>> compressed_array_with_filter = np.frombuffer(schunk.get_slice(), dtype=output_dtype)
        >>> f"Compressed data with prefilter applied (first 8 elements): {compressed_array_with_filter[:8]}"
        Compressed data with prefilter applied (first 8 elements): [-3.1415927  -2.1415927  -1.1415926  -0.14159265  0.8584073   1.8584074
         2.8584073   3.8584073 ]
        >>> schunk.remove_prefilter('prefilter')
        >>> schunk[:1000] = data
        >>> compressed_array_without_filter = np.frombuffer(schunk.get_slice(), dtype=dtype)
        >>> f"Compressed data without prefilter (first 8 elements): {compressed_array_without_filter[:8]}"
        Compressed data without prefilter (first 8 elements): [0. 1. 2. 3. 4. 5. 6. 7.]
        """
        return super().remove_prefilter(func_name)

    def __dealloc__(self):
        super().__dealloc__()


def open(
    urlpath: str | pathlib.Path | blosc2.URLPath, mode: str = "a", offset: int = 0, **kwargs: dict
) -> blosc2.SChunk | blosc2.NDArray | blosc2.C2Array | blosc2.LazyArray | blosc2.Proxy:
    """Open a persistent :ref:`SChunk`, :ref:`NDArray`, a remote :ref:`C2Array`
    or a :ref:`Proxy`

    See the `Notes` section for more info on opening `Proxy` objects.

    Parameters
    ----------
    urlpath: str | pathlib.Path | :ref:`URLPath`
        The path where the :ref:`SChunk` (or :ref:`NDArray`)
        is stored. If it is a remote array, a :ref:`URLPath` must be passed.
    mode: str, optional
        The open mode.
    offset: int, optional
        An offset in the file where super-chunk or array data is located
        (e.g. in a file containing several such objects).
    kwargs: dict, optional
        mmap_mode: The memory mapping mode.
        initial_mapping_size: The initial size of the memory mapping.
        cparams: dict
            A dictionary with the compression parameters, which are the same that can be
            used in the :func:`~blosc2.compress2` function.
            Typesize and blocksize cannot be changed.
        dparams: dict
            A dictionary with the decompression parameters, which are the same that can
            be used in the :func:`~blosc2.decompress2` function.

    Returns
    -------
    out: :ref:`SChunk`, :ref:`NDArray` or :ref:`C2Array`
        The SChunk or NDArray (if there is a "b2nd" metalayer")
        or the C2Array if :paramref:`urlpath` is a :ref:`blosc2.URLPath <URLPath>` instance.

    Notes
    -----
    * This is just a 'logical' open, so there is no `close()` counterpart because
      currently, there is no need for it.

    * If :paramref:`urlpath` is a :ref:`URLPath` instance, :paramref:`mode`
      must be 'r', :paramref:`offset` must be 0, and kwargs cannot be passed.

    * If the original object saved in :paramref:`urlpath` is a :ref:`Proxy`,
      this function will only return a :ref:`Proxy` if its source is a local
      :ref:`SChunk`, :ref:`NDArray` or a remote :ref:`C2Array`. Otherwise,
      it will return the Python-Blosc2 container used to cache the data which
      can be a :ref:`SChunk` or a :ref:`NDArray` and may not have all the data
      initialized (e.g. if the user has not accessed to it yet).

    * When opening a :ref:`LazyExpr` keep in mind the note above regarding operands.

    Examples
    --------
    >>> import blosc2
    >>> import numpy as np
    >>> import os
    >>> import tempfile
    >>> tmpdirname = tempfile.mkdtemp()
    >>> urlpath = os.path.join(tmpdirname, 'b2frame')
    >>> storage = blosc2.Storage(contiguous=True, urlpath=urlpath, mode="w")
    >>> nelem = 20 * 1000
    >>> nchunks = 5
    >>> chunksize = nelem * 4 // nchunks
    >>> data = np.arange(nelem, dtype="int32")
    >>> # Create SChunk and append data
    >>> schunk = blosc2.SChunk(chunksize=chunksize, data=data.tobytes(), storage=storage)
    >>> # Open SChunk
    >>> sc_open = blosc2.open(urlpath=urlpath)
    >>> for i in range(nchunks):
    ...     dest = np.empty(nelem // nchunks, dtype=data.dtype)
    ...     schunk.decompress_chunk(i, dest)
    ...     dest1 = np.empty(nelem // nchunks, dtype=data.dtype)
    ...     sc_open.decompress_chunk(i, dest1)
    ...     np.array_equal(dest, dest1)
    True
    True
    True
    True
    True

    To open the same schunk memory-mapped, we simply need to pass the `mmap_mode` parameter:

    >>> sc_open_mmap = blosc2.open(urlpath=urlpath, mmap_mode="r")
    >>> sc_open.nchunks == sc_open_mmap.nchunks
    True
    >>> all(sc_open.decompress_chunk(i, dest1) == sc_open_mmap.decompress_chunk(i, dest1) for i in range(nchunks))
    True
    """
    if isinstance(urlpath, blosc2.URLPath):
        if mode != "r" or offset != 0 or kwargs != {}:
            raise NotImplementedError(
                "Cannot open a C2Array with mode != 'r', or offset != 0 or some kwargs"
            )
        return blosc2.C2Array(urlpath.path, urlbase=urlpath.urlbase, auth_token=urlpath.auth_token)

    if isinstance(urlpath, pathlib.PurePath):
        urlpath = str(urlpath)
    if not os.path.exists(urlpath):
        raise FileNotFoundError(f"No such file or directory: {urlpath}")

    res = blosc2_ext.open(urlpath, mode, offset, **kwargs)

    meta = getattr(res, "schunk", res).meta
    if "proxy-source" in meta:
        proxy_src = meta["proxy-source"]
        if proxy_src["local_abspath"] is not None:
            src = blosc2.open(proxy_src["local_abspath"])
            return blosc2.Proxy(src, _cache=res)
        elif proxy_src["urlpath"] is not None:
            src = blosc2.C2Array(proxy_src["urlpath"][0], proxy_src["urlpath"][1], proxy_src["urlpath"][2])
            return blosc2.Proxy(src, _cache=res)
        elif not proxy_src["caterva2_env"]:
            raise RuntimeError("Could not find the source when opening a Proxy")

    if isinstance(res, blosc2.NDArray) and "LazyArray" in res.schunk.meta:
        return blosc2._open_lazyarray(res)
    else:
        return res



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Store metadata in persistent arrays

import numpy as np

import blosc2

shape = (128, 128)
urlpath = "ex_meta.b2nd"
dtype = np.complex128

# Create a numpy array
nparray = np.arange(int(np.prod(shape)), dtype=dtype).reshape(shape)

meta = {
    "m1": b"1111",
    "m2": b"2222",
}
# Create a NDArray from a numpy array (on disk)
a = blosc2.frombuffer(bytes(nparray), nparray.shape, urlpath=urlpath, mode="w", dtype=dtype, meta=meta)
print(a.info)

# Read a b2nd array from disk
b = blosc2.open(urlpath)

# Deal with meta
m1 = b.schunk.meta.get("m5", b"0000")
m2 = b.schunk.meta["m2"]
print("m1 meta:", m1)
print("m2 meta:", m2)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Examples of using the jit decorator with reductions
# You can find benchmarks for this example in the bench/ndarray directory

import numpy as np

import blosc2


# Example 1: Basic usage of the jit decorator with reduction
@blosc2.jit
def expr_jit(a, b, c):
    # This function computes a sum reduction along axis 1
    return np.sum(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=1)


# Create some sample data
a = blosc2.linspace(0, 1, 10 * 100, dtype="float32", shape=(10, 100))
b = blosc2.linspace(1, 2, 10 * 100, dtype="float32", shape=(10, 100))
c = blosc2.linspace(-10, 10, 10, dtype="float32", shape=(100,))

# Call the function with the jit decorator
result = expr_jit(a, b, c)
print("Example 1 result:", result)

# Example 2: Using the jit decorator with an out parameter for reduction
out = np.zeros((10,), dtype=np.int64)


@blosc2.jit
def expr_jit_out(a, b, c):
    # This function computes a sum reduction along axis 1 and stores the result in the 'out' array
    return np.sum(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=1, out=out)


# Call the function with the jit decorator and out parameter
result_out = expr_jit_out(a, b, c)
print("Example 2 result:", result_out)
print("Example 2 out array:", out)  # The 'out' array should now contain the same result

# Example 3: Using the jit decorator with additional keyword arguments for reduction
cparams = blosc2.CParams(clevel=1, codec=blosc2.Codec.LZ4, filters=[blosc2.Filter.BITSHUFFLE])
out_cparams = blosc2.zeros((10,), dtype=np.int64, cparams=cparams)


@blosc2.jit
def expr_jit_cparams(a, b, c):
    # This function computes a sum reduction along axis 1 with custom compression parameters
    return np.sum(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=1, out=out_cparams)


# Call the function with the jit decorator and custom parameters
result_cparams = expr_jit_cparams(a, b, c)
print("Example 3 result:", result_cparams[...])
print("Example 3 out array:", out_cparams[...])  # The 'out_cparams' array should now contain the same result



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

from dataclasses import asdict, replace

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("func", "op_dtype", "op2_dtype", "schunk_dtype", "offset"),
    [
        ("fill_f1", np.dtype(np.int32), None, np.dtype(np.int64), 0),
        ("fill_f1", np.dtype(np.int32), None, np.dtype(np.float32), 0),
        ("fill_f1", np.dtype(np.complex128), None, np.dtype(np.complex128), 0),
        ("fill_f2", np.dtype(np.float64), np.dtype(np.int32), np.dtype(np.float64), None),
        ("fill_f3", np.dtype("M8[D]"), None, np.dtype(np.bool_), None),
        ("fill_f4", np.dtype(np.float32), np.dtype(np.int32), np.dtype(np.float64), None),
    ],
)
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks", "contiguous", "urlpath", "nelem"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "nthreads": 1}, {"nthreads": 4}, 2, True, None, None),
        ({"nthreads": 1}, {"nthreads": 2}, 1, True, "test_fillers.b2frame", 1 * 20_000),
        (
            {"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 1},
            {"nthreads": 4},
            5,
            False,
            None,
            5 * 20_000,
        ),
        (
            {"codec": blosc2.Codec.LZ4HC, "nthreads": 1},
            {"nthreads": 1},
            3,
            False,
            "test_fillers.b2frame",
            None,
        ),
    ],
)
def test_fillers(  # noqa: C901
    contiguous, urlpath, cparams, dparams, nchunks, nelem, func, op_dtype, op2_dtype, schunk_dtype, offset
):
    blosc2.remove_urlpath(urlpath)

    chunk_len = 20_000
    cparams["typesize"] = schunk_dtype.itemsize

    schunk = blosc2.SChunk(
        chunksize=chunk_len * schunk_dtype.itemsize,
        contiguous=contiguous,
        urlpath=urlpath,
        cparams=cparams,
        dparams=dparams,
    )

    data = np.arange(0, chunk_len * nchunks, dtype=op_dtype)
    schunk_op = blosc2.SChunk(
        chunksize=chunk_len * op_dtype.itemsize, data=data, cparams={"typesize": op_dtype.itemsize}
    )
    res = np.empty(chunk_len * nchunks, dtype=schunk_dtype)
    if func == "fill_f1":

        @schunk.filler(((schunk_op, op_dtype),), schunk_dtype, nelem)
        def fill_f1(inputs_tuple, output, offset):
            for i in range(output.size):
                output[i] = offset + i

        fill_f1((data,), res, offset)

    elif func == "fill_f2":
        data2 = np.full(chunk_len * nchunks, 3, dtype=op2_dtype)
        schunk_op2 = blosc2.SChunk(
            chunksize=chunk_len * op2_dtype.itemsize, data=data2, cparams={"typesize": op2_dtype.itemsize}
        )

        @schunk.filler(((schunk_op, op_dtype), (schunk_op2, op2_dtype)), schunk_dtype, nelem)
        def fill_f2(inputs_tuple, output, offset):
            output[:] = inputs_tuple[0] * inputs_tuple[1]

        fill_f2((data, data2), res, offset)

    elif func == "fill_f3":

        @schunk.filler(((schunk_op, op_dtype),), schunk_dtype, nelem)
        def fill_f3(inputs_tuple, output, offset):
            output[:] = inputs_tuple[0] <= np.datetime64("1997-12-31")

        fill_f3((data,), res, offset)
    else:
        data2 = np.full(chunk_len * nchunks, 3, dtype=op2_dtype)

        @schunk.filler(((schunk_op, op_dtype), (data2, op2_dtype), (np.pi, np.float32)), schunk_dtype, nelem)
        def fill_f4(inputs_tuple, output, offset):
            output[:] = inputs_tuple[0] - inputs_tuple[1] * inputs_tuple[2]

        fill_f4((data, data2, np.pi), res, offset)

    new_cparams = replace(schunk.cparams, nthreads=2)
    schunk.cparams = new_cparams

    pre_data = np.empty(chunk_len * nchunks, dtype=schunk_dtype)
    schunk.get_slice(0, chunk_len * nchunks, out=pre_data)

    if "f" in schunk_dtype.str:
        assert np.allclose(pre_data, res)
    else:
        assert np.array_equal(pre_data, res)

    # Update a chunk
    chunk = np.full(chunk_len, 4, dtype=schunk_dtype)
    schunk[0:chunk_len] = chunk
    sl = np.empty(chunk_len, dtype=schunk_dtype)
    schunk.get_slice(0, chunk_len, sl)
    if "f" in schunk_dtype.str:
        assert np.allclose(chunk, sl)
    else:
        assert np.array_equal(chunk, sl)

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("func", "data_dtype", "schunk_dtype", "offset"),
    [
        ("pref1", np.dtype(np.int32), None, 0),
        ("pref1", np.dtype(np.int32), np.dtype(np.float32), 0),
        ("pref2", np.dtype(np.complex128), None, 0),
        ("pref2", np.dtype(np.float64), None, None),
        ("pref3", np.dtype("M8[D]"), np.dtype(np.int64), None),
    ],
)
@pytest.mark.parametrize(
    ("cparams", "dparams", "nchunks", "contiguous", "urlpath"),
    [
        ({"codec": blosc2.Codec.LZ4, "clevel": 6, "nthreads": 1}, {}, 2, True, None),
        ({"nthreads": 1}, {"nthreads": 2}, 1, True, "test_prefilters.b2frame"),
        ({"splitmode": blosc2.SplitMode.ALWAYS_SPLIT, "nthreads": 1}, {"nthreads": 4}, 5, False, None),
        ({"codec": blosc2.Codec.LZ4HC, "nthreads": 1}, {"nthreads": 4}, 3, False, "test_prefilters.b2frame"),
    ],
)
def test_prefilters(contiguous, urlpath, cparams, dparams, nchunks, func, data_dtype, schunk_dtype, offset):
    blosc2.remove_urlpath(urlpath)

    schunk_dtype = data_dtype if schunk_dtype is None else schunk_dtype
    chunk_len = 2_000
    data = np.arange(0, chunk_len * nchunks, dtype=data_dtype)
    cparams["typesize"] = schunk_dtype.itemsize
    schunk = blosc2.SChunk(
        chunksize=chunk_len * schunk_dtype.itemsize,
        contiguous=contiguous,
        urlpath=urlpath,
        cparams=cparams,
        dparams=dparams,
    )
    if func == "pref1":

        @schunk.prefilter(data_dtype, schunk_dtype)
        def pref1(input, output, offset):
            for i in range(input.size):
                output[i] = offset + i

    elif func == "pref2":

        @schunk.prefilter(data_dtype, schunk_dtype)
        def pref2(input, output, offset):
            output[:] = input - np.pi

    else:

        @schunk.prefilter(data_dtype, schunk_dtype)
        def pref3(input, output, offset):
            output[:] = input <= np.datetime64("1997-12-31")

    new_cparams = asdict(schunk.cparams)
    new_cparams["nthreads"] = 1
    schunk.cparams = blosc2.CParams(**new_cparams)

    schunk[: nchunks * chunk_len] = data
    post_data = np.empty(chunk_len * nchunks, dtype=schunk_dtype)
    schunk.get_slice(0, chunk_len * nchunks, out=post_data)

    res = np.empty(chunk_len * nchunks, dtype=schunk_dtype)
    locals()[func](data, res, offset)
    if "f" in data_dtype.str:
        assert np.allclose(post_data, res)
    else:
        assert np.array_equal(post_data, res)

    schunk.remove_prefilter(func)
    new_data = np.full(chunk_len, 5, dtype=schunk_dtype)
    schunk[:chunk_len] = new_data
    res = np.empty(chunk_len, dtype=schunk_dtype)
    schunk.get_slice(0, chunk_len, res)
    if "f" in data_dtype.str:
        assert np.allclose(new_data, res)
    else:
        assert np.array_equal(new_data, res)

    blosc2.remove_urlpath(urlpath)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################


import numpy as np
import pytest

import blosc2

random = np.random.default_rng()


@pytest.mark.parametrize("gil", [True, False])
@pytest.mark.parametrize(
    ("object", "codec"),
    [
        (random.integers(0, 10, 10), blosc2.Codec.LZ4),
        (np.arange(10), blosc2.Codec.BLOSCLZ),
        (random.integers(0, 1000, 1000, endpoint=True), blosc2.Codec.LZ4HC),
        (np.arange(45, dtype=np.float64), blosc2.Codec.ZLIB),
        (np.arange(50, dtype=np.int64), blosc2.Codec.ZSTD),
    ],
)
def test_decompress_numpy(object, codec, gil):
    blosc2.set_releasegil(gil)
    typesize = None if hasattr(object, "itemsize") else 1
    c = blosc2.compress(object, typesize=typesize, codec=codec)

    dest = bytearray(object)
    blosc2.decompress(c, dst=dest)
    assert dest == object.tobytes()

    dest2 = np.empty(object.shape, object.dtype)
    blosc2.decompress(c, dst=dest2)
    assert np.array_equal(dest2, object)

    dest3 = blosc2.decompress(c)
    assert dest3 == object.tobytes()

    dest4 = blosc2.decompress(c, as_bytearray=True)
    assert dest4 == object.tobytes()

    dest5 = np.empty(object.shape, object.dtype)
    blosc2.decompress(c, dst=memoryview(dest5))
    assert np.array_equal(dest5, object)


@pytest.mark.parametrize(
    ("object", "codec"),
    [
        (bytearray([0, 12, 24, 33]), blosc2.Codec.LZ4),
        (bytearray([2, 45, 6, 12, 78, 43, 23, 234]), blosc2.Codec.BLOSCLZ),
        (b"A string", blosc2.Codec.LZ4HC),
        (bytearray("Another string" * 100, encoding="utf-8"), blosc2.Codec.ZSTD),
    ],
)
def test_decompress(object, codec):
    c = blosc2.compress(object, typesize=1, codec=codec)

    dest = bytearray(object)
    blosc2.decompress(c, dst=dest)
    assert dest == object

    dest3 = blosc2.decompress(c)
    assert dest3 == object

    dest4 = blosc2.decompress(c, as_bytearray=True)
    assert dest4 == object

    dest5 = bytearray(object)
    blosc2.decompress(np.array([c]), dst=dest5)
    assert dest5 == object


@pytest.mark.parametrize(("object", "codec"), [(np.arange(0), blosc2.Codec.LZ4), (b"", blosc2.Codec.ZLIB)])
def test_raise_error(object, codec):
    c = blosc2.compress(object, codec=codec)

    dest = bytearray(object)
    with pytest.raises(ValueError):
        blosc2.decompress(c, dst=dest)

    dest3 = blosc2.decompress(c)
    if isinstance(object, bytes):
        assert dest3 == object
    else:
        assert dest3 == object.tobytes()

    dest4 = blosc2.decompress(c, as_bytearray=True)
    if isinstance(object, bytes):
        assert dest4 == object
    else:
        assert dest4 == object.tobytes()

    dest5 = bytearray(object)
    with pytest.raises(ValueError):
        blosc2.decompress(np.array([c]), dst=dest5)

    with pytest.raises(ValueError):
        blosc2.decompress(b"")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import math

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "chunks", "blocks", "dtype", "cparams", "urlpath", "contiguous", "meta"),
    [
        (
            (100, 1230),
            (200, 100),
            (55, 3),
            np.int32,
            {"codec": blosc2.Codec.ZSTD, "clevel": 4, "use_dict": 0, "nthreads": 1},
            None,
            True,
            None,
        ),
        (
            (23, 34),
            (10, 10),
            (10, 10),
            np.float64,
            {"codec": blosc2.Codec.BLOSCLZ, "clevel": 8, "use_dict": False, "nthreads": 2},
            "zeros.b2nd",
            True,
            {"abc": 123456789, "2": [0, 1, 23]},
        ),
        (
            (80, 51, 60),
            (20, 10, 33),
            (6, 6, 26),
            np.bool_,
            {"codec": blosc2.Codec.LZ4, "clevel": 5, "use_dict": 1, "nthreads": 2},
            None,
            False,
            {"abc": 123, "2": [0, 1, 24]},
        ),
        (
            (2**31 - 1,),
            (2**30,),
            None,
            np.float32,
            {"codec": blosc2.Codec.LZ4, "clevel": 5, "nthreads": 2},
            None,
            False,
            None,
        ),
    ],
)
def test_zeros(shape, chunks, blocks, dtype, cparams, urlpath, contiguous, meta):
    blosc2.remove_urlpath(urlpath)

    dtype = np.dtype(dtype)
    if math.prod(chunks) * dtype.itemsize > blosc2.MAX_BUFFERSIZE:
        with pytest.raises(RuntimeError):
            _ = blosc2.zeros(
                shape,
                chunks=chunks,
                blocks=blocks,
                dtype=dtype,
                cparams=cparams,
                urlpath=urlpath,
                contiguous=contiguous,
                meta=meta,
            )
        return
    else:
        a = blosc2.zeros(
            shape,
            chunks=chunks,
            blocks=blocks,
            dtype=dtype,
            cparams=cparams,
            urlpath=urlpath,
            contiguous=contiguous,
            meta=meta,
        )

    b = np.zeros(shape=shape, dtype=dtype)
    assert np.array_equal(a[:], b)

    if meta is not None:
        for metalayer in meta:
            m = a.schunk.meta[metalayer]
            assert m == meta[metalayer]

    blosc2.remove_urlpath(urlpath)


@pytest.mark.parametrize(
    ("shape", "dtype"),
    [
        (100, np.uint8),
        ((100, 1230), np.uint8),
        ((234, 125), np.int32),
        ((80, 51, 60), np.bool_),
        ((400, 399, 401), np.float64),
    ],
)
def test_zeros_minimal(shape, dtype):
    a = blosc2.zeros(shape, dtype=dtype)

    b = np.zeros(shape=shape, dtype=dtype)
    assert np.array_equal(a[:], b)

    dtype = np.dtype(dtype)
    assert shape in (a.shape, a.shape[0])
    assert a.chunks is not None
    assert a.blocks is not None
    assert all(c >= b for c, b in zip(a.chunks, a.blocks, strict=False))
    assert a.dtype == dtype
    assert a.schunk.typesize == dtype.itemsize



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2

argnames = "shape, chunks, blocks, slices, dtype"
argvalues = [
    ([456], [258], [73], slice(0, 1), np.int32),
    ([77, 134, 13], [31, 13, 5], [7, 8, 3], (slice(3, 7), slice(50, 100), 7), np.float64),
    ([12, 13, 14, 15, 16], [5, 5, 5, 5, 5], [2, 2, 2, 2, 2], (slice(1, 3), ..., slice(3, 6)), np.float32),
]


@pytest.mark.parametrize(argnames, argvalues)
def test_setitem(shape, chunks, blocks, slices, dtype):
    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=dtype).reshape(shape)
    a = blosc2.frombuffer(bytes(nparray), nparray.shape, dtype=dtype, chunks=chunks, blocks=blocks)

    # Python scalar
    nparray = a[...]
    a[slices] = 0
    nparray[slices] = 0
    np.testing.assert_almost_equal(a[...], nparray)

    # Object supporting the Buffer Protocol
    slice_shape = a[slices].shape
    val = np.ones(slice_shape, dtype=dtype)
    a[slices] = val
    nparray[slices] = val
    np.testing.assert_almost_equal(a[...], nparray)

    # blosc2.NDArray
    if np.prod(slice_shape) == 1 or len(slice_shape) != len(blocks):
        chunks = None
        blocks = None

    b = blosc2.full(slice_shape, fill_value=1234567, chunks=chunks, blocks=blocks, dtype=dtype)
    a[slices] = b
    nparray[slices] = b[...]
    np.testing.assert_almost_equal(a[...], nparray)


@pytest.mark.parametrize(
    ("shape", "slices"),
    [
        ([456], slice(0, 1)),
        ([77, 134, 13], (slice(3, 7), slice(50, 100), 7)),
        ([12, 13, 14, 15, 16], (slice(1, 3), ..., slice(3, 6))),
    ],
)
def test_setitem_different_dtype(shape, slices):
    size = int(np.prod(shape))
    nparray = np.arange(size, dtype=np.int32).reshape(shape)
    a = blosc2.empty(nparray.shape, dtype=np.float64)

    a[slices] = nparray[slices]
    nparray_ = nparray.astype(a.dtype)
    np.testing.assert_almost_equal(a[slices], nparray_[slices])


def test_ndfield():
    # Create a structured NumPy array
    shape = (50, 50)
    na = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
    nb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
    nsa = np.empty(shape, dtype=[("a", na.dtype), ("b", nb.dtype)])
    nsa["a"] = na
    nsa["b"] = nb
    sa = blosc2.asarray(nsa)

    # Check values
    assert np.allclose(sa["a"][:], na)
    assert np.allclose(sa["b"][:], nb)

    # Change values
    nsa["a"][:] = nsa["b"]
    sa["a"][:] = sa["b"]

    # Check values
    assert np.allclose(sa["a"][:], nsa["a"])
    assert np.allclose(sa["b"][:], nsa["b"])

    # Using NDField accessor
    nsa["b"][:] = 1
    fb = blosc2.NDField(sa, "b")
    fb[:] = blosc2.full(shape, fill_value=1, dtype=np.float64)
    assert np.allclose(sa["a"][:], nsa["a"])
    assert np.allclose(sa["b"][:], nsa["b"])
